'''
Stats.py
======

statistical functions
'''

import math
import numpy
import types

try:
    import scipy.interpolate
    HAS_SCIPY = True
except ImportError:
    HAS_SCIPY = False

class FDRResult:

    def __init__(self):
        pass

def computeQValues(pvalues, 
                   vlambda=None,
                   pi0_method="smoother", 
                   fdr_level=None, 
                   robust=False,
                   smooth_df = 3,
                   smooth_log_pi0 = False,
                   pi0 = None):
    """compute qvalues after the method by Storey et al. (2002)

    The python code derives from the R implementation at 
    http://genomics.princeton.edu/storeylab/qvalue/linux.html.
    """

    if min(pvalues) < 0 or max(pvalues) > 1:
        raise ValueError( "p-values out of range" )

    m = len(pvalues)
    pvalues = numpy.array( pvalues, dtype = numpy.float )

    if vlambda == None: vlambda = numpy.arange(0,0.95,0.05)


    if pi0 == None:
        if type(vlambda) == float:
            vlambda = (vlambda,)

        if len(vlambda) > 1 and len(vlambda) < 4:
            raise ValueError(" if length of vlambda greater than 1, you need at least 4 values." )

        if len(vlambda) > 1 and (min(vlambda) < 0 or max(vlambda) >= 1):
            raise ValueError( "vlambda must be within [0, 1).")

        # estimate pi0
        if len(vlambda)==1: 
            vlambda = vlambda[0]
            if  vlambda < 0 or vlambda >=1 :
                raise ValueError( "vlambda must be within [0, 1).")

            pi0 = numpy.mean( [ x >= vlambda for x in pvalues ] ) / (1.0 - vlambda)
            pi0 = min(pi0, 1.0)
        else:

            pi0 = numpy.zeros( len(vlambda), numpy.float )

            for i in range( len(vlambda) ):
                pi0[i] = numpy.mean( [x >= vlambda[i] for x in pvalues ]) / (1.0 -vlambda[i] )

            if pi0_method=="smoother":
                if smooth_log_pi0: pi0 = numpy.log(pi0)
                if HAS_SCIPY:
                    tck = scipy.interpolate.splrep( vlambda, pi0, k = smooth_df, s = 10000 )
                    pi0 = scipy.interpolate.splev( max(vlambda), tck )
                else:
                    raise ImportError( "pi0_method smoother requires scipy" )

                if smooth_log_pi0: pi0 = numpy.exp(pi0)
                
            elif pi0_method=="bootstrap":
                minpi0 = min(pi0)

                mse = numpy.zeros( len(vlambda), numpy.float )
                pi0_boot = numpy.zeros( len(vlambda), numpy.float )

                for i in xrange(100):
                    # sample pvalues
                    idx_boot = numpy.random.random_integers( 0, m-1, m) 
                    pvalues_boot = pvalues[idx_boot]

                    for x in xrange( len(vlambda )):
                        # compute number of pvalues larger than lambda[x]
                        pi0_boot[x] = numpy.mean( pvalues_boot > vlambda[x]) / (1.0 - vlambda[x]) 
                    mse += (pi0_boot - minpi0) ** 2
                pi0 = min( pi0[mse==min(mse)] )
            else:
                raise ValueError( "'pi0_method' must be one of 'smoother' or 'bootstrap'.")

            pi0 = min(pi0,1.0)
    
    if pi0 <= 0:
        raise ValueError( "The estimated pi0 <= 0 (%f). Check that you have valid p-values or use another vlambda method." %  pi0)

    if fdr_level != None and (fdr_level <= 0 or fdr_level > 1):
        raise ValueError( "'fdr_level' must be within (0, 1].")

    # compute qvalues

    idx = numpy.argsort( pvalues )
    # monotonically decreasing bins, so that bins[i-1] > x >=  bins[i]
    bins = numpy.unique( pvalues )[::-1]

    # v[i] = number of observations less than or equal to pvalue[i]
    # could this be done more elegantly?
    val2bin = len(bins) - numpy.digitize( pvalues, bins )
    v = numpy.zeros( m, dtype = numpy.int )
    lastbin = None
    for x in xrange( m-1, -1, -1 ):
        bin = val2bin[idx[x]]
        if bin != lastbin: c = x
        v[idx[x]] = c+1
        lastbin = bin

    qvalues = pvalues * pi0 * m / v
    if robust:
        qvalues /= ( 1.0 - ( 1.0 - pvalues)**m )

    # bound qvalues by 1 and make them monotonic
    qvalues[idx[m-1]] = min(qvalues[idx[m-1]],1.0)
    for i in xrange(m-2,-1,-1):
        qvalues[idx[i]] = min(min(qvalues[idx[i]],qvalues[idx[i+1]]),1.0)

    # fill result
    result = FDRResult()
    result.qvalues = qvalues

    if fdr_level != None:
        result.passed = [ x <= fdr_level for x in result.qvalues ]
    else:
        result.passed = [ False for x in result.qvalues ]
        
    result.pvalues = pvalues
    result.pi0 = pi0
    result.vlambda = vlambda
    result.fdr_level = fdr_level

    return result



## {{{ http://code.activestate.com/recipes/511478/ (r1)
def percentile(N, percent, key=lambda x:x):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values. Note N MUST BE already sorted.
    @parameter percent - a float value from 0.0 to 1.0.
    @parameter key - optional key function to compute value from each element of N.

    @return - the percentile of the values
    """
    if len(N) == 0: return None
    k = (len(N)-1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return key(N[int(k)])
    d0 = key(N[int(f)]) * (c-k)
    d1 = key(N[int(c)]) * (k-f)
    return d0+d1

###################################################################
###################################################################
###################################################################
## adjust P-Value
###################################################################
def adjustPValues( pvalues, method = 'fdr', n = None ):
    '''returns an array of adjusted pvalues
    
    Reimplementation of p.adjust in the R package.

    p: numeric vector of p-values (possibly with 'NA's).  Any other
    R is coerced by 'as.numeric'.

    method: correction method. Valid values are:

    n: number of comparisons, must be at least 'length(p)'; only set
    this (to non-default) when you know what you are doing
    
    For more information, see the documentation of the
    p.adjust method in R.
    '''

    if n == None: n = len(pvalues)

    if method == "fdr": method = "BH"

    # optional, remove NA values
    p = numpy.array( pvalues, dtype = numpy.float )
    lp = len(p)
    
    assert n <= lp

    if n <= 1:
        return p
    if n == 2 and method == "hommel":
        method = "hochberg"

    if method == "bonferroni":
        p0 = n * p 
    elif method == "holm":
        i = numpy.arange( lp )
        o = numpy.argsort( p )
        ro = numpy.argsort( o )
        m = numpy.maximum.accumulate( (n - i) * p[o] )
        p0 = m[ro]
    elif method == "hommel":
        raise NotImplementedError( "hommel method not fully implemented" )
    elif method == "hochberg":
        i = numpy.arange( 0, lp)[::-1] 
        o = numpy.argsort(1-p)
        ro = numpy.argsort( o )
        m = numpy.minimum.accumulate( (n - i) * p[o] )
        p0 = m[ro]
    elif method == "BH":
        i = numpy.arange( 1, lp + 1)[::-1] 
        o = numpy.argsort(1-p)
        ro = numpy.argsort( o )
        m = numpy.minimum.accumulate( float(n) / i * p[o] )
        p0 = m[ro]
    elif method == "BY":
        i = numpy.arange( 1, lp + 1)[::-1] 
        o = numpy.argsort(1-p)
        ro = numpy.argsort( o )
        q = numpy.sum( 1.0 / numpy.arange( 1, n + 1) )
        m = numpy.minimum.accumulate( q * float(n) / i * p[o])
        p0 = m[ro]
    elif method == "none":
        p0 = p

    return numpy.minimum( p0, numpy.ones( len(p0) ) )

class Result(object):
    '''allow both member and dictionary access.'''
    slots=("_data")
    def __init__(self):
        object.__setattr__(self, "_data", dict())
    def fromR( self, take, r_result ):
        '''convert from an *r_result* dictionary using map *take*.

        *take* is a list of tuples mapping a field to the corresponding
        field in *r_result*.
        '''
        for x,y in take:
            if y:
                self._data[x] = r_result.rx(y)[0][0]
            else:
                self._data[x] = r_result.rx(x)[0][0]

            # if y:
            #     self._data[x] = r_result[y]
            # else:
            #     self._data[x] = r_result[x]

        return self
    def __getattr__(self, key):
        if not key.startswith("_"):
            try: return object.__getattribute__(self,"_data")[key]
            except KeyError: pass
        return getattr( self._data, key )
    def keys(self): return self._data.keys()
    def values(self): return self._data.values()
    def __len__(self): return self._data.__len__()
    def __str__(self):
        return str(self._data)
    def __contains__(self,key):
        return key in self._data
    def __getitem__(self, key ):
        return self._data[key]
    def __delitem__(self, key ):
        del self._data[key]
    def __setitem__(self, key, value ):
        self._data[key] = value
    def __setattr__(self, key, value):
        if not key.startswith("_"):
            self._data[key] = value
        else:
            object.__setattr__(self,key,value)

class Summary( Result ):
    """a collection of distributional parameters. Available properties
    are:

    mean, median, min, max, samplestd, sum, counts
    """

    fields = ("nval", "min", "max", "mean", "median", "stddev", "sum", "q1", "q3")

    def __init__(self, values = None, 
                 format = "%6.4f", mode="float",
                 allow_empty = True ):

        Result.__init__(self)
        self._format = format
        self._mode = mode

        # note that this determintes the order of the fields at output
        self.counts, self.min, self.max, self.mean, self.median, self.samplestd, self.sum, self.q1, self.q3 = \
                    (0, 0, 0, 0, 0, 0, 0, 0, 0)
        
        if values != None:

            values = [x for x in values if x != None ]

            if len(values) == 0:
                if allow_empty: return
                else: raise ValueError( "no data for statistics" )

            # convert
            self._nerrors = 0
            if type(values[0]) not in (types.IntType, types.FloatType):
                n = []
                for x in values:
                    try:
                        n.append( float(x) )
                    except ValueError:
                        self._nerrors += 1
            else:
                n = values

            ## use a non-sort algorithm?
            n.sort()
            if len(n):
                self.q1 = n[len(n) / 4]
                self.q3 = n[len(n) * 3 / 4]
            else:
                self.q1 = self.q3 = 0 

            self.counts = len(n)
            self.min = min(n)
            self.max = max(n)
            self.mean = numpy.mean( n )
            self.median = numpy.median( n )
            self.samplestd = numpy.std( n )
            self.sum = reduce( lambda x, y: x+y, n )

    def getHeaders( self ):
        """returns header of column separated values."""
        return self.fields

    def getHeader( self ):
        """returns header of column separated values."""
        return "\t".join( self.getHeaders())

    def __str__( self ):
        """return string representation of data."""
        
        if self._mode == "int":
            format_vals = "%i"
            format_median = "%.1f"
        else:
            format_vals = self._format
            format_median = self._format

        return "\t".join( ( "%i" % self.counts,
                            format_vals % self.min,
                            format_vals % self.max,
                            self._format % self.mean,
                            format_median % self.median,
                            self._format % self.samplestd,                                      
                            format_vals % self.sum,
                            format_vals % self.q1,
                            format_vals % self.q3,                            
                            ) )



################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id$
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
Experiment.py - record keeping of experiments
=============================================

Module for record keeping of experiments. This module
is imported by most CGAT scripts. It provides convenience
methods for

   * argument parsing
   * record keeping (logging)
   * benchmarking

See :doc:`../scripts/script_template` on how to use this module.

API
---
'''

import string,re,sys,time,inspect,os,optparse,logging,random, collections, types, subprocess, gzip


class DefaultOptions:
    stdlog = sys.stdout
    stdout = sys.stdout
    stderr = sys.stderr

global_starting_time = time.time()
global_options = DefaultOptions()
global_args    = None
# import hashlib
# global_id = hashlib.md5(time.asctime(time.localtime(time.time()))).hexdigest()
import uuid
global_id      = uuid.uuid4()
global_benchmark = collections.defaultdict( int )

def GetHeader():
    """return a header string with command line options and
    timestamp"""
    system, host, release, version, machine = os.uname()
    return "# output generated by %s\n# job started at %s on %s -- %s\n# pid: %i, system: %s %s %s %s" %\
           (" ".join(sys.argv),
            time.asctime(time.localtime(time.time())),
            host,
            global_id,
            os.getpid(),
            system, release, version, machine)

def GetParams( options = None):
    """return a string containing script parameters.
    Parameters are all variables that start with "param_".
    """
    result = []
    if options:
        members = options.__dict__
        for k, v in sorted(members.items()):
            result.append("# %-40s: %s" % (k, str(v).encode("string_escape")))
    else:
        vars = inspect.currentframe().f_back.f_locals
        for var in filter(lambda x: re.match("param_",x), vars.keys()):
            result.append("# %-40s: %s" % (var, str(vars[var]).encode("string_escape")))

    if result:
        return "\n".join(result)
    else:
        return "# no parameters."


def GetFooter():
    """return a header string with command line options and
    timestamp."""
    return "# job finished in %i seconds at %s -- %s -- %s" %\
           (time.time() - global_starting_time,
            time.asctime(time.localtime(time.time())),
            " ".join( map( lambda x: "%5.2f" % x, os.times()[:4])),
            global_id)

def Start( parser = None,
           argv = sys.argv,
           quiet = False,
           add_csv_options = False,
           add_mysql_options = False,
           add_psql_options = False,
           add_pipe_options = True,
           add_cluster_options = False,
           add_output_options = False ):
    """set up an experiment.

    returns a tuple containing (options, args).

    The options class is extended with a logger module.
    """
    
    if not parser:
        parser = optparse.OptionParser( version = "%prog version: $Id: Experiment.py 2803 2009-10-22 13:41:24Z andreas $" )

    global global_options, global_args, global_starting_time

    global_starting_time = time.time()

    parser.add_option("-v", "--verbose", dest="loglevel", type="int",
                      help="loglevel [%default]. The higher, the more output." )

    parser.add_option( "--timeit", dest='timeit_file', type="string",
                       help="store timeing information in file [%default]." )        
    parser.add_option( "--timeit-name", dest='timeit_name', type="string",
                       help="name in timing file for this class of jobs [%default]." )
    parser.add_option( "--timeit-header", dest='timeit_header', action="store_true",
                       help="add header for timing information [%default]." )        
                                     
    if quiet:
        parser.set_defaults( loglevel = 0 )
    else:
        parser.set_defaults( loglevel = 1 )

    parser.set_defaults(
        timeit_file = None,
        timeit_name = 'all',
        timeit_header = None,
        )

    if add_csv_options:
        parser.add_option("--dialect", dest="csv_dialect", type="string",
                          help="csv dialect to use [%default]." )

        parser.set_defaults(
            csv_dialect = "excel-tab",
            csv_lineterminator = "\n",
            )

    if add_psql_options:
        parser.add_option("-C", "--connection", dest="psql_connection", type="string",
                          help="psql connection string [%default]."  )
        parser.add_option("-U", "--user", dest="user", type="string",
                          help="database user name [%default]."  )
        
        parser.set_defaults( psql_connection = "db:andreas" )
        parser.set_defaults( user = "" )
        
    if add_cluster_options:
        parser.add_option( "--use-cluster", dest="use_cluster", action = "store_true",
                          help="use cluster [%default]."  )
        parser.add_option( "--cluster-priority", dest="cluster_priority", type="int",
                          help="set job priority on cluster [%default]."  )
        parser.add_option( "--cluster-queue", dest="cluster_queue", type="string",
                          help="set cluster queue [%default]."  )
        parser.add_option( "--cluster-num-jobs", dest="cluster_num_jobs", type="int",
                          help="number of jobs to submit to the queue execute in parallel [%default]."  )
        parser.add_option( "--cluster-options", dest="cluster_options", type="string",
                          help="additional options for cluster jobs, passed on to qrsh [%default]."  )
        
        parser.set_defaults( use_cluster = False,
                             cluster_queue = "medium_jobs.q",
                             cluster_priority = -10,
                             cluster_num_jobs = 100,
                             cluster_options = "")

    if add_output_options:
        parser.add_option( "-P", "--output-filename-pattern", dest="output_filename_pattern", type="string" ,
                           help="OUTPUT filename pattern for various methods [%default].")

        parser.add_option( "-F", "--force", dest="output_force", action="store_true",
                           help="force over-writing of existing files." )

        parser.set_defaults( output_filename_pattern = "%s",
                             output_force = False )

    if add_pipe_options:
        parser.add_option("-I", "--stdin", dest="stdin", type="string",
                          help="file to read stdin from [default = stdin].",
                          metavar = "FILE"  )
        parser.add_option("-L", "--log", dest="stdlog", type="string",
                          help="file with logging information [default = stdout].",
                          metavar = "FILE"  )
        parser.add_option("-E", "--error", dest="stderr", type="string",
                          help="file with error information [default = stderr].",
                          metavar = "FILE"  )
        parser.add_option("-S", "--stdout", dest="stdout", type="string",
                          help="file where output is to go [default = stdout].",
                          metavar = "FILE"  )
        
        parser.set_defaults( stderr = sys.stderr )
        parser.set_defaults( stdout = sys.stdout )
        parser.set_defaults( stdlog = sys.stdout )
        parser.set_defaults( stdin = sys.stdin )
        
    if add_mysql_options:
        parser.add_option("-H", "--host", dest="host", type="string",
                          help="mysql host [%default]."  )
        parser.add_option("-D", "--database", dest="database", type="string",
                          help="mysql database [%default]."  )
        parser.add_option("-U", "--user", dest="user", type="string",
                          help="mysql username [%default]."  )
        parser.add_option("-P", "--password", dest="password", type="string",
                          help="mysql password [%default]."  )
        parser.add_option("-O", "--port", dest="port", type="int",
                          help="mysql port [%default]."  )
        
        parser.set_defaults( host = "db",
                             port = 3306,
                             user = "",
                             password = "",
                             database = "" )

    (global_options, global_args) = parser.parse_args( argv[1:] )

    if add_pipe_options:
        if global_options.stdout != sys.stdout: 
            global_options.stdout = open(global_options.stdout, "w")
        if global_options.stderr != sys.stderr:
            if global_options.stderr == "stderr":
                global_options.stderr = global_options.stderr
            else:
                global_options.stderr = open(global_options.stderr, "w")
        if global_options.stdlog != sys.stdout:
            global_options.stdlog = open(global_options.stdlog, "a")            
        if global_options.stdin != sys.stdin: 
            global_options.stdin = open(global_options.stdin, "r")
    else:
        global_options.stderr = sys.stderr
        global_options.stdout = sys.stdout
        global_options.stdlog = sys.stdout
        global_options.stdin = sys.stdin
    
    if global_options.loglevel >= 1:
        global_options.stdlog.write(GetHeader() + "\n" )
        global_options.stdlog.write(GetParams( global_options) + "\n")
        global_options.stdlog.flush()

    ## configure logging
    ## map from 0-10 to logging scale
    ## 0: quiet
    ## 1: little verbositiy
    ## >1: increased verbosity
    if global_options.loglevel == 0:
        lvl = logging.ERROR
    elif global_options.loglevel == 1:
        lvl = logging.INFO
    else:
        lvl = logging.DEBUG

    if global_options.stdout == global_options.stdlog:
        logging.basicConfig(
            level=lvl,
            format='# %(asctime)s %(levelname)s %(message)s',
            stream = global_options.stdlog )
    else:
        logging.basicConfig(
            level=lvl,
            format='%(asctime)s %(levelname)s %(message)s',
            stream = global_options.stdlog )
        
    return global_options, global_args
    
def Stop():
    """stop the experiment."""

    if global_options.loglevel >= 1 and global_benchmark:
        t = time.time() - global_starting_time
        global_options.stdlog.write( "######### Time spent in benchmarked functions ###################\n" )
        global_options.stdlog.write( "# function\tseconds\tpercent\n" )
        for key, value in global_benchmark.items():
            global_options.stdlog.write( "# %s\t%6i\t%5.2f%%\n" % (key, value, (100.0 * float(value) / t)))
        global_options.stdlog.write( "#################################################################\n" )

    if global_options.loglevel >= 1:
        global_options.stdlog.write(GetFooter() + "\n")

    # close files
    if global_options.stdout != sys.stdout: global_options.stdout.close()
    # do not close log, otherwise the following error occurs:
    # Error in sys.exitfunc:
    # Traceback (most recent call last):
    #   File "/net/cpp-group/server/lib/python2.6/atexit.py", line 24, in _run_exitfuncs
    #     func(*targs, **kargs)
    #   File "/net/cpp-group/server/lib/python2.6/logging/__init__.py", line 1472, in shutdown
    #     h.flush()
    #   File "/net/cpp-group/server/lib/python2.6/logging/__init__.py", line 740, in flush
    #     self.stream.flush()
    # ValueError: I/O operation on closed file
    # if global_options.stdlog != sys.stdout: global_options.stdlog.close()
    if global_options.stderr != sys.stderr: global_options.stderr.close()

    if global_options.timeit_file:

        outfile = open(global_options.timeit_file, "a")

        if global_options.timeit_header:
            outfile.write( "\t".join( ("name", "wall", "user", "sys", "cuser", "csys",
                                       "host", "system", "release", "machine",
                                       "start", "end", "path", "cmd" ) ) + "\n" )

        csystem, host, release, version, machine = map(str, os.uname())
        uusr, usys, c_usr, c_sys = map( lambda x: "%5.2f" % x, os.times()[:4])
        t_end = time.time()
        c_wall = "%5.2f" % (t_end - global_starting_time)
        
        if sys.argv[0] == "run.py":
            cmd = global_args[0]
            if len(global_args) > 1:
                cmd += " '" + "' '".join(global_args[1:]) + "'" 
        else:
            cmd = sys.argv[0]

        result = "\t".join( (global_options.timeit_name,
                             c_wall, uusr, usys, c_usr, c_sys,
                             host, csystem, release, machine,
                             time.asctime(time.localtime( global_starting_time )),
                             time.asctime(time.localtime( t_end )),
                             os.path.abspath( os.getcwd() ),
                             cmd ) ) + "\n"

        outfile.write( result )
        outfile.close()

def benchmark(func):
    """decorator collecting wall clock time spent in decorated method."""

    def wrapper(*arg):
        t1 = time.time()
        res = func(*arg)
        t2 = time.time()
        key = "%s:%i" % (func.func_name, func.func_code.co_firstlineno)
        global_benchmark[key] += t2-t1
        global_options.stdlog.write( '## benchmark: %s completed in %6.4f s\n' % (key, (t2-t1)))
        global_options.stdlog.flush()
        return res
    return wrapper

# there are differences whether you cache a function or
# an objects method
def cachedmethod(function):
    '''decorator for caching a method.'''
    return Memoize(function)

class Memoize(object):
    def __init__(self, fn):
        self.cache={}
        self.fn=fn

    def __get__(self, instance, cls=None):
        self.instance = instance
        return self

    def __call__(self,*args):
        if self.cache.has_key(args):
            return self.cache[args]
        else:
            object = self.cache[args] = self.fn(self.instance, *args)
            return object

def log( loglevel, message ):
    """log message at loglevel."""
    logging.log( loglevel, message )

def info( message ):
    '''log information message, see the :mod:`logging` module'''
    logging.info( message )

def warning( message ):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning( message )

def warn( message ):
    '''log warning message, see the :mod:`logging` module'''
    logging.warning( message )

def debug( message ):
    '''log debugging message, see the :mod:`logging` module'''
    logging.debug( message )

def error( message ):
    '''log error message, see the :mod:`logging` module'''
    logging.error( message )
        
def critical( message):
    '''log critical message, see the :mod:`logging` module'''
    logging.critical( message )

def getOutputFile( section ):
    '''return filename to write to.'''
    return re.sub( "%s", section, global_options.output_filename_pattern )
    
def openOutputFile( section, mode = "w" ):
    """open file for writing substituting section in the
    output_pattern (if defined).

    If the filename ends with ".gz", the output is opened
    as a gzip'ed file.
    """

    fn = getOutputFile( section )
    try:
        if fn == "-":
            return global_options.stdout
        else:
            if not global_options.output_force and os.path.exists( fn ):
                raise OSError( "file %s already exists, use --force to overwrite existing files." % fn )
            if fn.endswith(".gz"):
                return gzip.open( fn, mode )
            else:
                return open( fn, mode )
    except AttributeError:        
        return global_options.stdout


class Counter(object):
    '''a counter class.

    The counter acts both as a dictionary and
    a object permitting attribute access.

    Counts are automatically initialized to 0.
    
    Instantiate and use like this::

       c = Counter()
       c.input += 1
       c.output += 2
       c["skipped"] += 1

       print str(c)
    '''

    __slots__=["_counts"]
    def __init__(self):
        """Store data returned by function."""
        object.__setattr__( self, "_counts", collections.defaultdict( int ) )

    def __setitem__(self, key, value ):
        self._counts[key] = value
    def __getitem__(self, key):
        return self._counts[key]
    def __getattr__(self, name):
        return self._counts[name]
    def __setattr__(self, name, value):
        self._counts[name] = value
    def __str__(self):
        return ", ".join( "%s=%i" % x for x in self._counts.iteritems() )
    def __iadd__(self, other ):
        try:
            for key,val in other.iteritems():
                self._counts[key] += val
        except:
            raise TypeError( "unknown type" )
        return self
    def iteritems(self):
        return self._counts.iteritems()

class Experiment:
    
    mShortOptions = ""
    mLongOptions  = []

    mLogLevel = 0
    mTest = 0
    mDatabaseName = None

    mName = sys.argv[0]


    def __init__(self ):

        # process command-line arguments
        (self.mOptlist, self.mArgs) = self.ParseCommandLine()

        # set options now
        self.ProcessOptions(self.mOptlist)

    def DumpParameters( self ):
        """dump parameters of this object. All parameters start with a lower-case m."""

        members = self.__dict__

	print "#--------------------------------------------------------------------------------------------"
        print "#" + string.join(sys.argv)
        print "# pid: %i, system:" % os.getpid(), string.join(os.uname(), ",") 
	print "#--------------------------------------------------------------------------------------------"
        print "# Parameters for instance of <" + self.mName + "> on " + time.asctime(time.localtime(time.time()))

        member_keys = list(members.keys())
        member_keys.sort()
        for member in member_keys:
            if member[0] == 'm':
                print "# %-40s:" % member, members[member]

	print "#--------------------------------------------------------------------------------------------"
        sys.stdout.flush()

    #-----------------------------> Control functions <--------------------------------

    ##------------------------------------------------------------------------------------
    def ProcessOptions( self, optlist ):
        """Sets options in this module. Please overload as necessary."""

        for o,a in optlist:
            if o in ( "-V", "--Verbose" ):
                self.mLogLevel = string.atoi(a)
            elif o in ( "-T", "--test") :
                self.mTest = 1

    ##------------------------------------------------------------------------------------        
    def ProcessArguments( self, args ):
        """Perform actions as given in command line arguments."""

        if self.mLogLevel >= 1:
            self.DumpParameters()
            
        for arg in args:
            if arg[-1] == ")":
                statement = "self.%s" % arg
            else:
                statement = "self.%s()" % arg
            exec statement

            if self.mLogLevel >= 1:
                print "--------------------------------------------------------------------------------------------"
                print statement + " finished at " + time.asctime(time.localtime(time.time()))
                print "--------------------------------------------------------------------------------------------"

    ##------------------------------------------------------------------------------------            
    def ParseCommandLine( self ):
        """Call subroutine with command line arguments."""

        self.mShortOptions = self.mShortOptions + "V:D:T"
        self.mLongOptions.append( "Verbose=" )
        self.mLongOptions.append( "Database=" )
        self.mLongOptions.append( "Test" )

        try:
            optlist, args = getopt.getopt(sys.argv[1:],
                                          self.mShortOptions,
                                          self.mLongOptions)
        except getopt.error, msg:
            self.PrintUsage()
            print msg
            sys.exit(2)

        return optlist, args

    #--------------------------------------------------------------------------------        
    def Process( self ):
        self.ProcessArguments(self.mArgs)

    #--------------------------------------------------------------------------------
    def PrintUsage( self ):
        """print usage information."""

        print "# valid short options are:", self.mShortOptions
        print "# valid long options are:", str(self.mLongOptions)
    
def run( cmd ):
    '''executed a command line cmd.
    
    raises OSError if process failed or was terminated.
    '''

    retcode = subprocess.call( cmd, shell=True)
    if retcode < 0:
        raise OSError( "process was terminated by signal %i" % -retcode )
    return retcode



import unittest
import random, tempfile, shutil, os, re, gzip, sys, subprocess, collections
import numpy, math

from GatSegmentList import SegmentList
import GatEngine
import gat
import gat.IO
import gat.Stats

class Options: pass

class TestRunning( unittest.TestCase ):
    '''test running gat.

    Gat is being run in single-threaded and multi-threaded modes and the output
    is compared to reference gat results.

    The reference data has been created with the following commands::
       
        gat-run.py --segments=data/segments_multiple.bed.gz --annotations=data/annotations.bed.gz --workspace=data/workspace.bed.gz --num-samples=1000 > data/output_multiple.tsv
        gat-run.py --segments=data/segments_single.bed.gz --annotations=data/annotations.bed.gz --workspace=data/workspace.bed.gz --num-samples=1000 > data/output_single.tsv
    '''

    sample_size = 1000
    
    # similarity threshold for float comparisons
    threshold = 0.20

    filename_segments = [ 'data/segments_single.bed.gz' ]
    filename_annotations = [ 'data/annotations.bed.gz' ]
    filename_workspace = [ 'data/workspace.bed.gz' ]

    max_percent_difference = 10
    mean_percent_difference = 5

    def setUp( self ):

        parser = gat.buildParser()

        options, args = parser.parse_args([])

        options.segment_files = self.filename_segments
        options.annotation_files = self.filename_annotations
        options.workspace_files = self.filename_workspace

        self.segments, self.annotations, workspaces, isochores = gat.IO.buildSegments( options )
        self.workspace = gat.IO.applyIsochores( self.segments, 
                                                self.annotations,
                                                workspaces,
                                                options,
                                                isochores )
        
        self.sampler = GatEngine.SamplerAnnotator( bucket_size = 1, nbuckets = 100000 )

        self.counters = [GatEngine.CounterNucleotideOverlap()]
        self.workspace_generator = GatEngine.UnconditionalWorkspace()
        
        
        self.reference_data = gat.IO.readAnnotatorResults( 'data/output_single.tsv' )

    def compare( self, annotator_results ):

        self.assertEqual( len(annotator_results), 
                          len(self.segments) * len(self.annotations)
                          )

        self.assertEqual( len(annotator_results),
                          len(self.reference_data ) )

        self.assertEqual( sorted([x.track for x in annotator_results]),
                          sorted([x.track for x in self.reference_data]) )

        self.assertEqual( sorted([x.annotation for x in annotator_results]),
                          sorted([x.annotation for x in self.reference_data]) )
        
        # sort according to reference
        positions = dict( [ (y,x) for x,y in enumerate( [(x.track, x.annotation) for x in self.reference_data] )] )
        annotator_results.sort( key = lambda x: positions[(x.track,x.annotation)] )

        def _output_stats( s ):
            return ",".join( [ "%s=%s" % (x,y) for x,y in zip( summary.getHeaders(), 
                                                             str(summary).split("\t")) ] )
        for attr in ("expected", "fold", "pvalue" ):
            
            this = [ getattr(x,attr) for x in annotator_results ]
            ref = [ getattr(x,attr) for x in self.reference_data ]
        
            percent_differences = [ 100.0 * (x - y) / y for x,y in zip( this, ref ) ]
            summary = gat.Stats.Summary( percent_differences )
            mm = max(abs(summary.min), summary.max)
            self.assertLess( mm, self.max_percent_difference, 
                             "attribute '%s': maximum percent difference %f > %f: %s" \
                                 % (attr, 
                                    mm, self.max_percent_difference, 
                                 _output_stats( summary) ))

            self.assertLess( summary.mean, self.mean_percent_difference, 
                             "attribute '%s': mean percent difference %f > %f: %s" \
                                 % (attr, 
                                    summary.mean, self.max_percent_difference, 
                                 _output_stats( summary) ))
            

        # observed should be identical
        for ref in self.reference_data:
            this = [x for x in annotator_results if x.track == ref.track and x.annotation == ref.annotation ][0]
            self.assertEqual( ref.observed, this.observed )

    def testNoMultiprocessing( self ):

        annotator_results = gat.run( self.segments,
                                     self.annotations,
                                     self.workspace,
                                     self.sampler,
                                     self.counters,
                                     workspace_generator = self.workspace_generator,
                                     num_samples = self.sample_size )

        self.compare( annotator_results )

    def testOneCPU( self ):

        annotator_results = gat.run( self.segments,
                                     self.annotations,
                                     self.workspace,
                                     self.sampler,
                                     self.counters,
                                     workspace_generator = self.workspace_generator,
                                     num_samples = self.sample_size,
                                     num_threads = 1)
        
        self.compare( annotator_results )

    def testTwoCPU( self ):
        annotator_results = gat.run( self.segments,
                                     self.annotations,
                                     self.workspace,
                                     self.sampler,
                                     self.counters,
                                     workspace_generator = self.workspace_generator,
                                     num_samples = self.sample_size,
                                     num_threads = 2)
        
        self.compare( annotator_results )

if __name__ == '__main__':
    unittest.main()



cdef extern from "string.h":
    ctypedef int size_t
    void *memcpy(void *dest, void *src, size_t n)
    void *memmove(void *dest, void *src, size_t n)
    char *strtok_r(char *str, char *delim, char **saveptr)
    char *strncpy(char *dest, char *src, size_t n)
    void *memchr(void *s, int c, size_t n)

cdef extern from "stdlib.h":
    void free(void *)
    void *malloc(size_t)
    void *calloc(size_t,size_t)
    void *realloc(void *,size_t)
    int c_abs "abs" (int)
    int atoi( char *nptr)
    long atol( char *nptr)
    double atof( char *nptr)
    void qsort(void *base, size_t nmemb, size_t size,
               int(*compar)(void *, void *))
    int rand()
    int rand_r(unsigned int *seedp)
    void srand(unsigned int seed)


cdef extern from "stdint.h":
    ctypedef int int64_t
    ctypedef int int32_t
    ctypedef int uint32_t
    ctypedef int uint8_t
    ctypedef int uint64_t

cdef extern from "stdio.h":
    ctypedef struct FILE
    # check 32/64 bit compatibility for large files
    ctypedef long off_t
    ctypedef long off64_t
    ctypedef struct fpos_t:
        pass
    cdef int SEEK_SET
    cdef int SEEK_CUR
    cdef int SEEK_END
    FILE *fopen(char *,char *)
    int fclose(FILE *)
    int feof(FILE *)
    int sscanf(char *str,char *fmt,...)
    int sprintf(char *str,char *fmt,...)
    int fprintf(FILE *ifile,char *fmt,...)
    int ferror(FILE *stream)
    int fseeko(FILE *stream, off_t offset, int whence)
    off_t ftello(FILE *stream)
    size_t fwrite( void *ptr,
                   size_t size,
                   size_t nmemb,
                   FILE *stream)
    size_t fread(void *ptr,
                 size_t size,
                 size_t nmemb,
                 FILE *stream)
    char *fgets(char *str,int size,FILE *ifile)
    int fgetpos(FILE *stream, fpos_t *pos)
    int fsetpos(FILE *stream, fpos_t *pos)
    int printf(char *format, ...)

cdef extern from "math.h":
    double floor(double x)


cdef extern from "errno.h":
    cdef int errno

cdef extern from "sys/stat.h":
    ctypedef int mode_t 
    cdef int S_IRUSR
    cdef int S_IWUSR

cdef extern from "fcntl.h":
    cdef int O_CREAT
    cdef int O_RDWR
    cdef int O_RDONLY

cdef extern from "sys/mman.h":

    int shm_open(char *name, int oflag, mode_t mode)
    int shm_unlink(char *name)
    void *mmap(void *addr, size_t length, int prot, int flags,
    	            int fd, off_t offset)
    void *mmap64(void *addr, size_t length, int prot, int flags,
    	            int fd, off64_t offset)
    int munmap(void *addr, size_t length)
    cdef void * MAP_FAILED
    cdef int PROT_READ
    cdef int PROT_WRITE
    cdef int MAP_SHARED    

cdef extern from "unistd.h":
    int ftruncate(int fd, off_t length)

cdef extern from "Python.h":
    ctypedef struct FILE
    FILE* PyFile_AsFile(object)
    char *fgets(char *str, int size, FILE *ifile)
    int feof(FILE *stream)
    size_t strlen(char *s)
    size_t getline(char **lineptr, size_t *n, FILE *stream)
    char *strstr(char *, char *)
    char *strchr(char *string, int c)
    int fileno(FILE *stream)

# Next, enter the builtin file class into the namespace:
cdef extern from "fileobject.h":
    ctypedef class __builtin__.file [object PyFileObject]:
        pass

cdef extern from "gat_utils.h":
    long searchsorted(void * base,
                       size_t nmemb,
                       size_t size,
                       void * target,
                       int(*compar)(void *, void *))

    long searchargsorted(void * base,
                         int * sorted,
                         size_t nmemb,
                         size_t size,
                         void * target,
                         int(*compar)(void *, void *))

    int toCompressedFile( unsigned char *, size_t, FILE *)
    int fromCompressedFile( unsigned char *, size_t, FILE *)

cdef class IntervalContainer:
    cdef int shared_fd     	   
    cdef str shared_fn
    cdef intervals
    cdef str name
 
cdef class IntervalDictionary(IntervalContainer):
    pass

cdef class IntervalCollection(IntervalContainer):
    pass


    



cdef extern from "string.h":
    ctypedef int size_t
    void *memcpy(void *dest, void *src, size_t n)
    void *memmove(void *dest, void *src, size_t n)
    char *strtok_r(char *str, char *delim, char **saveptr)
    char *strncpy(char *dest, char *src, size_t n)
    void *memchr(void *s, int c, size_t n)
    char * strerror_r( int, char *, int )
    

cdef extern from "stdlib.h":
    void free(void *)
    void *malloc(size_t)
    void *calloc(size_t,size_t)
    void *realloc(void *,size_t)
    int c_abs "abs" (int)
    int atoi( char *nptr)
    long atol( char *nptr)
    double atof( char *nptr)
    void qsort(void *base, size_t nmemb, size_t size,
               int(*compar)(void *, void *))
    int rand()
    int rand_r(unsigned int *seedp)
    void srand(unsigned int seed)

cdef extern from "stdint.h":
    ctypedef int int64_t
    ctypedef int int32_t
    ctypedef int uint32_t
    ctypedef int uint8_t
    ctypedef int uint64_t

cdef extern from "stdio.h":
    ctypedef struct FILE
    # check 32/64 bit compatibility for large files
    ctypedef long off_t
    ctypedef long off64_t
    ctypedef struct fpos_t:
        pass
    cdef int SEEK_SET
    cdef int SEEK_CUR
    cdef int SEEK_END
    FILE *fopen(char *,char *)
    int fclose(FILE *)
    int feof(FILE *)
    int sscanf(char *str,char *fmt,...)
    int sprintf(char *str,char *fmt,...)
    int fprintf(FILE *ifile,char *fmt,...)
    int ferror(FILE *stream)
    int fseeko(FILE *stream, off_t offset, int whence)
    off_t ftello(FILE *stream)
    void perror(char *)
    size_t fwrite( void *ptr,
                   size_t size,
                   size_t nmemb,
                   FILE *stream)
    size_t fread(void *ptr,
                 size_t size,
                 size_t nmemb,
                 FILE *stream)
    char *fgets(char *str,int size,FILE *ifile)
    int fgetpos(FILE *stream, fpos_t *pos)
    int fsetpos(FILE *stream, fpos_t *pos)
    int printf(char *format, ...)

cdef extern from "errno.h":
    cdef int errno

cdef extern from "sys/stat.h":
    ctypedef int mode_t 
    cdef int S_IRUSR
    cdef int S_IWUSR

cdef extern from "fcntl.h":
    cdef int O_CREAT
    cdef int O_RDWR
    cdef int O_RDONLY

cdef extern from "sys/mman.h":

    int shm_open(char *name, int oflag, mode_t mode)
    int shm_unlink(char *name)
    void *mmap(void *addr, size_t length, int prot, int flags,
    	            int fd, off_t offset)
    void *mmap64(void *addr, size_t length, int prot, int flags,
    	            int fd, off64_t offset)
    int munmap(void *addr, size_t length)
    cdef void * MAP_FAILED
    cdef int PROT_READ
    cdef int PROT_WRITE
    cdef int MAP_SHARED    

cdef extern from "unistd.h":
    int ftruncate(int fd, off_t length)

cdef extern from "math.h":
    double floor(double x)

cdef extern from "Python.h":
    ctypedef struct FILE
    FILE* PyFile_AsFile(object)
    char *fgets(char *str, int size, FILE *ifile)
    int feof(FILE *stream)
    size_t strlen(char *s)
    size_t getline(char **lineptr, size_t *n, FILE *stream)
    char *strstr(char *, char *)
    char *strchr(char *string, int c)
    int fileno(FILE *stream)

# Next, enter the builtin file class into the namespace:
cdef extern from "fileobject.h":
    ctypedef class __builtin__.file [object PyFileObject]:
        pass

cdef extern from "gat_utils.h":
    long searchsorted(void * base,
                       size_t nmemb,
                       size_t size,
                       void * target,
                       int(*compar)(void *, void *))

    long searchargsorted(void * base,
                         int * sorted,
                         size_t nmemb,
                         size_t size,
                         void * target,
                         int(*compar)(void *, void *))

    int toCompressedFile( unsigned char *, size_t, FILE *)
    int fromCompressedFile( unsigned char *, size_t, FILE *)

#####################################################
#####################################################
## type definitions
#####################################################
# type for positions
ctypedef unsigned int Position
# type for position difference - can be negative
ctypedef int PositionDifference

# A segment (without strand)
cdef struct Segment:
    Position start
    Position end

#####################################################
#####################################################
## definition of segmentlist
#####################################################
cdef class SegmentList:

    cdef Segment * segments
    cdef size_t nsegments
    cdef size_t allocated
    cdef int flag
    cdef int chunk_size
    cdef int shared_fd
    cdef key

    # C and Python methods
    cpdef share( self, key )
    cpdef sort( self )
    cpdef SegmentList extend( self, SegmentList other )
    cpdef add( self, Position start, Position end )
    cpdef trim_ends( self, Position pos, Position size, int forward  )
    cpdef trim( self, Position pos, Position size )
    cpdef normalize( self )
    cpdef merge( self, PositionDifference distance  )
    cpdef check( self )
    cpdef Position getRandomPosition( self )
    cpdef Position overlapWithRange( self, Position start, Position end )
    cpdef SegmentList getOverlappingSegmentsWithRange( self, Position start, Position end )
    cpdef Position overlapWithSegments( self, SegmentList other )
    cpdef Position intersectionWithSegments( self, SegmentList other, mode = * )
    cpdef SegmentList getFilledSegmentsFromStart( self, Position start, PositionDifference remainder )
    cpdef SegmentList getFilledSegmentsFromEnd( self, Position end, PositionDifference remainder )
    cpdef SegmentList filter( self, SegmentList other )
    cpdef SegmentList intersect( self, SegmentList other )
    cpdef SegmentList subtract( self, SegmentList other )
    cpdef extend_segments( self, int extension )
    cpdef expand_segments( self, double expansion )
    cpdef Position sum( self )
    cpdef Position max( self )
    cpdef Position min( self )
    cpdef clear( self )
    cpdef summarize( self )

    # C only methods
    cdef _add( self, Segment segment )
    cdef _resize( self, int nsegments )
    cdef insert( self, int idx, Segment seg )
    cdef int _getInsertionPoint( self, Segment other )
    cdef Position overlap( self, Segment other )
    cdef SegmentList getOverlappingSegments( self, Segment other )
    cdef SegmentList truncate( self, Segment other )
    cdef off_t toMMAP( self, void *, int, off_t )
    cdef void fromMMAP( self )




import os, sys, re, optparse, collections, types, gzip, pprint, time
import numpy

import gat.Bed as Bed
import gat.IOTools as IOTools
import gat.Experiment as E
import gat.Stats as Stats

import GatEngine

import multiprocessing.pool
import multiprocessing

def readFromBedOld( filenames, name = "track" ):
    '''read Segment Lists from one or more bed files.
 
    Segment lists are grouped by *contig* and *track*.
    
    If no track is given, the *name* attribute is taken.
    '''

    segment_lists = collections.defaultdict( lambda: collections.defaultdict(SegmentList))

    if name == "track": f = lambda x: x.mTrack["name"]
    elif name == "name": f = lambda x: x.mFields[0]
    else: raise ValueError("unknown name: '%s'" %name )

    for filename in filenames:
        infile = IOTools.openFile( filename, "r")
        for bed in Bed.iterator( infile ):
            try:
                name = f(bed)
            except TypeError:
                name = "default"
            segment_lists[name][bed.contig].add( bed.start, bed.end )

    return segment_lists

def buildParser( usage = None ):
    '''return gat command line parser.
    '''

    parser = optparse.OptionParser( version = "%prog version: $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = usage )

    parser.add_option("-a", "--annotation-file", "--annotations", dest="annotation_files", type="string", action="append",
                      help="filename with annotations [default=%default]."  )

    parser.add_option("-s", "--segment-file", "--segments", dest="segment_files", type="string", action="append",
                      help="filename with segments. Also accepts a glob in parentheses [default=%default]."  )

    parser.add_option("-w", "--workspace-file", "--workspace", dest="workspace_files", type="string", action="append",
                      help="filename with workspace segments. Also accepts a glob in parentheses [default=%default]."  )

    parser.add_option("-i", "--isochore-file", "--isochores", dest="isochore_files", type="string", action="append",
                      help="filename with isochore segments. Also accepts a glob in parentheses [default=%default]."  )

    parser.add_option("-l", "--sample-file", dest="sample_files", type="string", action="append",
                      help="filename with sample files. Start processing from samples [default=%default]."  )

    parser.add_option("-c", "--counter", dest="counters", type="choice", action="append",
                      choices=("nucleotide-overlap", 
                               "nucleotide-density",
                               "segment-overlap", 
                               "segment-midoverlap",
                               "annotations-overlap", 
                               "annotations-midoverlap" ),
                      help="quantity to use for estimating enrichment [default=%default]."  )

    parser.add_option("-m", "--sampler", dest="sampler", type="choice",
                      choices=("annotator", 
                               "segments",
                               "shift",
                               "local-permutation",
                               "global-permutation",
                               "uniform",
                               "brute-force"),
                      help="quantity to test [default=%default]."  )

    parser.add_option("-n", "--num-samples", dest="num_samples", type="int", 
                      help="number of samples to compute [default=%default]."  )

    parser.add_option("-e", "--cache", dest="cache", type="string", 
                      help="filename for caching samples [default=%default]."  )

    parser.add_option("-o", "--order", dest="output_order", type="choice",
                      choices = ( "track", "annotation", "fold", "pvalue", "qvalue" ),
                      help="order results in output by fold, track, etc. [default=%default]."  )

    parser.add_option("-p", "--pvalue-method", dest="pvalue_method", type="choice",
                      choices = ( "empirical", "norm", ),
                      help="type of pvalue reported [default=%default]."  )

    parser.add_option("-q", "--qvalue-method", dest="qvalue_method", type="choice",
                      choices = ( "storey", "BH", "bonferroni", "holm", "hommel", "hochberg", "BY", "none" ),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default]."  )

    parser.add_option("-t", "--num-threads", dest="num_threads", type="int",
                      help = "number of threads to use for sampling [default=%default]" )

    parser.add_option( "--qvalue-lambda", dest="qvalue_lambda", type="float",
                      help="fdr computation: lambda [default=%default]."  )

    parser.add_option( "--qvalue-pi0-method", dest="qvalue_pi0_method", type="choice",
                       choices = ("smoother", "bootstrap" ),
                       help="fdr computation: method for estimating pi0 [default=%default]."  )

    parser.add_option( "--input-counts-file", dest="input_filename_counts", type="string", 
                      help="start processing from counts - no segments required [default=%default]."  )

    parser.add_option( "--input-results-file", dest="input_filename_results", type="string", 
                      help="start processing from results - no segments required [default=%default]."  )

    parser.add_option( "--output-tables-pattern", dest="output_tables_pattern", type="string", 
                      help="output pattern for result tables. Used if there are multiple counters used [default=%default]."  )

    parser.add_option( "--output-counts-pattern", dest="output_counts_pattern", type="string", 
                      help="output pattern for counts [default=%default]."  )

    parser.add_option( "--output-plots-pattern", dest="output_plots_pattern", type="string", 
                       help="output pattern for plots [default=%default]" )

    parser.add_option( "--output-samples-pattern", dest="output_samples_pattern", type="string", 
                       help="output pattern for samples. Samples are stored in bed format, one for "
                            " each segment [default=%default]" )

    parser.add_option( "--output-stats", dest="output_stats", type="choice", action="append",
                       choices = ( "all", 
                                   "annotations", "segments", 
                                   "workspaces", "isochores",
                                   "overlap", 
                                   "sample",
                                   "segment_metrics",
                                   "sample_metrics"),
                       help="output overlap summary stats [default=%default]."  )

    parser.add_option( "--output-bed", dest="output_bed", type="choice", action="append",
                       choices = ( "all", 
                                   "annotations", "segments", 
                                   "workspaces", "isochores",
                                   "overlap" ),
                       help="output bed files [default=%default]."  )
    
    parser.add_option( "--descriptions", dest="input_filename_descriptions", type="string", 
                       help="filename mapping annotation terms to descriptions. "
                            " if given, the output table will contain additional columns "
                            " [default=%default]" )

    parser.add_option( "--bucket-size", dest="bucket_size", type="int", 
                       help="size of a bin for histogram of segment lengths. If 0, it will be automatically"
                            " scaled to fit nbuckets [default=%default]" )

    parser.add_option( "--nbuckets", dest="nbuckets", type="int", 
                       help="number of bins for histogram of segment lengths [default=%default]" )

    parser.add_option( "--ignore-segment-tracks", dest="ignore_segment_tracks", action="store_true", 
                       help="ignore segment tracks - all segments belong to one track [default=%default]" )

    parser.add_option( "--enable-split-tracks", dest="enable_split_tracks", action="store_true", 
                       help="permit the same track to be in multiple files [default=%default]" )

    parser.add_option( "--conditional", dest="conditional", type="choice",
                       choices = ( "unconditional", "annotation-centered", "segment-centered", "cooccurance" ),
                       help="conditional workspace creation [default=%default]"
                       " cooccurance - compute enrichment only within workspace segments that contain both segments "
                       " and annotations"
                       " annotation-centered - workspace centered around annotations. See --conditional-extension"
                       " segment-centered - workspace centered around segments. See --conditional-extension" )

    parser.add_option( "--conditional-extension", dest="conditional_extension", type="int",
                      help="if workspace is created conditional, extend by this amount (in bp) [default=%default]."  )

    parser.add_option( "--conditional-expansion", dest="conditional_expansion", type="float",
                      help="if workspace is created conditional, expand by this amount (ratio) [default=%default]."  )

    parser.add_option( "--restrict-workspace", dest="restrict_workspace", action="store_true", 
                       help="restrict workspace to those segments that contain both track"
                       " and annotations [default=%default]" )

    parser.add_option( "--truncate-workspace-to-annotations", dest="truncate_workspace_to_annotations", action="store_true", 
                       help="truncate workspace with annotations [default=%default]" )

    parser.add_option( "--truncate-segments-to-workspace", dest = "truncate_segments_to_workspace", action="store_true",
                       help = "truncate segments to workspace before sampling [default=%default]" )

    parser.add_option( "--shift-extension", dest="shift_extension", type="float",
                      help="if the sampling method is 'shift', create a segment of size # anound the segment"
                       " to determine the size of the region for shifthing [default=%default]."  )

    parser.add_option( "--shift-expansion", dest="shift_expansion", type="float",
                      help="if the sampling method is 'shift', multiply each segment by # "
                           " to determine the size of the region for shifthing [default=%default]."  )

    parser.add_option( "--pseudo-count", dest="pseudo_count", type="float",
                      help="pseudo count. The pseudo count is added to both the observed and expected overlap. "
                       " Using a pseudo-count avoids gat reporting fold changes of 0 [default=%default]."  )

    parser.add_option( "--null", dest="null", type="string",
                      help="null hypothesis. The default is to test categories for enrichment/depletion. "
                           " If a filename with gat output is given, gat will test for the difference "
                           " in fold change between the segments supplied and in the other file [default=%default]."  )

    parser.set_defaults(
        annotation_files = [],
        segment_files = [],
        workspace_files = [],  
        sample_files = [],
        num_samples = 1000,
        nbuckets = 100000,
        bucket_size = 0,
        counters = [],
        output_stats = [],
        output_bed = [],
        output_order = "fold",
        cache = None,
        input_filename_counts = None,
        input_filename_results = None,
        pvalue_method = "empirical",
        output_tables_pattern = "%s.tsv.gz",
        output_counts_pattern = None,
        output_plots_pattern = None,
        output_samples_pattern = None,
        qvalue_method = "BH",
        qvalue_lambda = None,
        qvalue_pi0_method = "smoother",
        sampler = "annotator",
        ignore_segment_tracks = False,
        input_filename_descriptions = None,
        conditional = "unconditional",
        conditional_extension = None,
        conditional_expansion = None,
        restrict_workspace = False,
        truncate_segments_to_workspace = False,
        truncate_workspace_to_annotations = False,
        enable_split_tracks = False,
        shift_expansion = 2.0,
        shift_extension = 0,
        # pseudo count for fold change computation to avoid 0 fc
        pseudo_count = 1.0,
        null = "default",
        num_threads = 0,
        )

    return parser
    

def iterator_results( annotator_results ):
    '''iterate over all results.'''
    for k1, v1 in annotator_results.iteritems():
        for k2, v2 in v1.iteritems():
            yield v2

class DummyAnnotatorResult:

    format_observed = "%i"
    format_expected = "%6.4f"
    format_fold = "%6.4f"
    format_pvalue = "%6.4e"

    def __init__( self ):
        pass

    @classmethod
    def _fromLine( cls, line ):
        x = cls()
        data = line[:-1].split("\t")
        x.track, x.annotation = data[:2]
        x.counter = "na"
        x.observed, x.expected, x.lower95, x.upper95, x.stddev, x.fold, x.l2fold, x.pvalue, x.qvalue = \
            map(float, data[2:11] )
        if len(data) > 11:
            (track_nsegments,
             track_size,
             track_density,
             annotation_nsegments,
             annotation_size,
             annotation_density,
             overlap_nsegments,
             overlap_size,
             overlap_density,
             percent_overlap_nsegments_track,
             percent_overlap_size_track,
             percent_overlap_nsegments_annotation,
             percent_overlap_size_annotation) = map(float, data[11:24])
            

        return x

    def __str__(self):
        return "\t".join( (self.track,
                           self.annotation,
                           self.format_observed % self.observed,
                           self.format_expected % self.expected,
                           self.format_expected % self.lower95,
                           self.format_expected % self.upper95,
                           self.format_expected % self.stddev,
                           self.format_fold % self.fold,
                           self.format_pvalue % self.pvalue,
                           self.format_pvalue % self.qvalue ) )


WorkData = collections.namedtuple( "WorkData", 
                                   "track sample_id sampler segments "
                                   "annotations contig_annotations "
                                   "workspace contig_workspace "
                                   "counters " )

def computeSample( args ):
    '''compute a single sample.
    '''

    workdata, samples_outfile, metrics_outfile, lock = args

    ( track,
      sample_id,
      sampler, 
      segs, 
      annotations, 
      contig_annotations,
      workspace, 
      contig_workspace,
      counters ) = workdata

    # E.debug("track=%s, sample=%s - started" % (track, str(sample_id)))

    counts = Experiment.Counter()

    sample_id = str(sample_id)
    counts_per_isochore = collections.defaultdict( list )

    outf_samples = samples_outfile

    if samples_outfile: 
        if lock:
            lock.acquire()
            outf_samples = IOTools.openFile( samples_outfile, "a" )

        samples_outfile.write("track name=%s\n" % sample_id)

        if lock:
            outf_samples.close()
            lock.release()

    sample = GatEngine.IntervalDictionary()

    for isochore in segs.keys():

        counts.pairs += 1

        # skip empty isochores
        if workspace[isochore].isEmpty or segs[isochore].isEmpty: 
            counts.skipped += 1
            continue

        counts.sampled += 1
        r = sampler.sample( segs[isochore], workspace[isochore] )
   
        # TODO : activate
        # self.outputSampleStats( sample_id, isochore, r )

        sample.add( isochore, r )

        # save sample
        if samples_outfile: 
            if lock:
                lock.acquire()
                outf_samples = IOTools.openFile( samples_outfile, "a" )

            for start, end in r:
                outf_samples.write( "%s\t%i\t%i\n" % (isochore, start, end))
                
            if lock:
                outf_samples.close()
                lock.release()

    # re-combine isochores
    # adjacent intervals are merged.
    sample.fromIsochores()

    if metrics_outfile:
        if lock:
            lock.acquire()
            outf = IOTools.openFile( metrics_outfile, "a")
        else:
            outf = metrics_outfile

        IO.outputMetrics( outf, sample, workspace, track, sample_id )
        
        if lock:
            outf.close()
            lock.release()

    counts_per_track = [ collections.defaultdict(float) for x in counters ]
    # compute counts for each counter
    for counter_id, counter in enumerate(counters):
        # TODO: choose aggregator
        for annotation in annotations.tracks:
            counts_per_track[counter_id][annotation] = sum( [
                    counter( sample[contig],
                             contig_annotations[annotation][contig],
                             contig_workspace[contig])
                    for contig in sample.keys() ] )

    # E.debug("track=%s, sample=%s - completed" % (track,str(sample_id )))

    return counts_per_track

class UnconditionalSampler:

    def __init__(self,         
                 num_samples, samples, 
                 samples_outfile, 
                 sampler,
                 workspace_generator, 
                 counters,
                 outfiles,
                 num_threads = 1 ):
        self.num_samples = num_samples
        self.samples = samples
        self.samples_outfile = samples_outfile
        self.sampler = sampler
        self.workspace_generator = workspace_generator
        self.counters = counters
        
        self.outfile_sample_stats = outfiles.get( "sample_stats", None )
        self.outfile_sample_metrics = outfiles.get( "sample_metrics", None )

        if self.outfile_sample_stats:
            E.debug( "sample stats go to %s" % outfile_sample_stats)
            self.outfile_sample_stats.write( "sample\tisochore\tnsegments\tnnucleotides\tmean\tstd\tmin\tq1\tmedian\tq3\tmax\n" )

        self.last_sample_id = None
        self.all_lengths = []
        self.num_threads = num_threads

    def outputSampleStats( self, sample_id, isochore, sample ):

        def _write( sample_id, isochore, lengths ):
            if self.outfile_sample_stats:
                self.outfile_sample_stats.write( "\t".join( map(str, (\
                                sample_id,
                                isochore,
                                len(lengths),
                                numpy.sum( lengths ),
                                numpy.mean( lengths ),
                                numpy.std( lengths ),
                                min( lengths ),
                                Stats.percentile( lengths, 0.25 ),
                                numpy.median( lengths ),
                                Stats.percentile( lengths, 0.75 ),
                                max(lengths) ) )) + "\n" )

            
        if sample_id != self.last_sample_id:
            if self.last_sample_id != None:
                _write( self.last_sample_id, "all", numpy.sort(numpy.array(self.all_lengths)) )
            self.all_lengths = []
            self.last_sample_id = sample_id

            
        if sample_id:
            l = sample.asLengths()
            self.all_lengths.extend( l  )
            _write( sample_id, isochore, numpy.sort( numpy.array( l )))

    def computeSamples( self, work, report_interval = 100 ):
        '''compute samples according to work.

        returns a list of results.
        '''
        n = len(work)
        
        E.debug( 'sampling will work on %i items' % n)

        results = []

        if self.num_threads == 0:
            for i, w in enumerate(work):
                r = computeSample( (w, self.samples_outfile, self.outfile_sample_metrics, None)  )
                if i % report_interval == 0:
                    E.info( "%i/%i done (%5.2f)" % (i, n, 100.0 * i / n ))
                results.append( r )
        else:
            E.info("generating processpool with %i threads for %i items" % (self.num_threads, len(work) ))
            
            manager = multiprocessing.Manager()

            lock = manager.Lock()

            pool = multiprocessing.Pool( self.num_threads )

            # use file names - not files when multiprocessing
            samples_outfile, metrics_outfile = None, None
            if self.samples_outfile:
                samples_outfile = self.samples_outfile.name
                self.samples_outfile.flush()
            if self.outfile_sample_metrics:
                metrics_outfile = self.outfile_sample_metrics.name
                self.outfile_sample_metrics.flush()

            ww = [ (w, samples_outfile, metrics_outfile, lock ) for w in work ]

            for i, r in enumerate(pool.imap_unordered(computeSample, ww )):
                if i % report_interval == 0:
                    E.info( "%i/%i done (%5.2f)" % (i, n, 100.0 * i / n ))
                results.append( r )

            pool.close()
            pool.join()

        return results

    def sample( self, track, counts, counters, segs, annotations, workspace, outfiles ):
        '''sample and return counts.

        Return a list of counted results for each counter.
        '''

        E.info( "performing unconditional sampling" )
        counts_per_track = [ collections.defaultdict( list ) for x in counters ]

        # rebuild non-isochore annotations and workspace
        contig_annotations = annotations.clone()
        contig_annotations.fromIsochores()
        contig_annotations.setName( "contig_" + annotations.getName())

        contig_workspace = workspace.clone()
        contig_workspace.fromIsochores()

        E.info( "workspace without conditioning: %i segments, %i nucleotides" % \
                    (workspace.counts(),
                     workspace.sum() ) )

        temp_segs, _, temp_workspace = self.workspace_generator( segs, None, workspace )

        E.info( "workspace after conditioning: %i segments, %i nucleotides" % \
                     (workspace.counts(),
                      workspace.sum() ) )

        if workspace.sum() == 0:
            E.warn( "empty workspace - no computation performed" )
            return counts_per_track

        work = [ WorkData( track, 
                           x, 
                           self.sampler,
                           temp_segs, 
                           annotations,
                           contig_annotations,
                           temp_workspace,
                           contig_workspace,
                           counters, 
                           ) for x in range(self.num_samples) ]

        if self.num_threads > 0:
            E.info("setting up shared data for multi-processing")
            annotations.share()
            contig_annotations.share()
            contig_workspace.share( "contig_workspace" )
            temp_segs.share( "generated_segments" )
            temp_workspace.share( "generated_workspace" )

        E.info( "sampling started" )
        results = self.computeSamples( work )
        E.info( "sampling completed" )

        if self.num_threads > 0:
            E.info("retrieving private data")
            annotations.unshare()
            contig_annotations.unshare()
            contig_workspace.unshare()
            temp_segs.unshare()
            temp_workspace.unshare()

        # collate results
        for result in results:
            for counter_id, counter in enumerate(counters):
                for annotation in annotations.tracks:
                    counts_per_track[counter_id][annotation].append(
                        result[counter_id][annotation] )

        self.outputSampleStats( None, "", [] )

        return counts_per_track

class ConditionalSampler( UnconditionalSampler ):
    
    def sample( self, track, counts, counters, segs, annotations, workspace, outfiles ):
        '''conditional sampling - sample using only those 
        segments that contain both a segment and an annotation.

        return dictionary with counts per track
        '''

        E.info( "performing conditional sampling" )
        counts_per_track = [ collections.defaultdict( list ) for x in counters ]

        # rebuild non-isochore annotations and workspace
        contig_annotations = annotations.clone()
        contig_annotations.fromIsochores()
        contig_annotations.setName( "contig_" + annotations.getName() )

        contig_workspace = workspace.clone()
        contig_workspace.fromIsochores()

        E.info("setting up shared data for multi-processing")
        annotations.share()
        contig_annotations.share()
        contig_workspace.share( "contig_workspace" )

        E.info( "workspace without conditioning: %i segments, %i nucleotides" % \
                     (workspace.counts(),
                      workspace.sum() ) )

        if workspace.sum() == 0:
            E.warn( "empty workspace - no computation performed" )
            return counts_per_track

        # compute samples conditionally - need to proceed by annotation
        for annoid, annotation in enumerate(annotations.tracks):

            annos = annotations[annotation]

            temp_segs, temp_annotations, temp_workspace = self.workspace_generator( segs, annos, workspace )
            
            # set up sharing
            temp_segs.share( "generated_segments" )
            temp_workspace.share( "generated_workspace" )

            E.info( "workspace for annotation %s: %i segments, %i nucleotides" % \
                        (annotation,
                         temp_workspace.counts(),
                         temp_workspace.sum() ) )

            work = [ WorkData( '_'.join( (track,anno_id)),
                               x,
                               self.sampler,
                               temp_segs, 
                               annotations,
                               contig_annotations,
                               temp_workspace,
                               contig_workspace,
                               counters, 
                               ) for x in range(self.num_samples) ]
            
            E.info( "sampling for annotation '%s' started" % annotation)
            results = self.computeSamples( work )
            E.info( "sampling for annotation '%s' completed" % annotation)

            for result in results:
                for counter_id, counter in enumerate(counters):
                    counts_per_track[counter_id][annotation].append( result[counter_id][annotation] )

        return counts_per_track

def run( segments, 
         annotations, 
         workspace, 
         sampler, 
         counters,
         workspace_generator,
         **kwargs ):
    '''run an enrichment analysis.

    segments: an IntervalCollection
    workspace: an IntervalCollection
    annotations: an IntervalCollection

    kwargs recognized are:

    cache 
       filename of cache

    num_samples
       number of samples to compute

    output_counts_pattern
       output counts to filename 

    output_samples_pattern
       if given, output samles to these files, one per segment

    sample_files
       if given, read samples from these files.

    fdr
       method to compute qvalues

    outfiles
       dictionary of optional additional output files.

    pseudo_count
       pseudo_count to add to observed and expected values

    reference
       data with reference observed and expected values.
    '''

    ## get arguments
    num_samples = kwargs.get( "num_samples", 10000 )
    cache = kwargs.get( "cache", None )
    output_counts_pattern = kwargs.get( "output_counts_pattern", None )
    sample_files = kwargs.get( "sample_files", [] )
    pseudo_count = kwargs.get( "pseudo_count", 1.0 )
    reference = kwargs.get( "reference", None )
    output_samples_pattern = kwargs.get( "output_samples_pattern", None )
    outfiles = kwargs.get( "outfiles", {} )
    num_threads = kwargs.get( "num_threads", 0 )

    ##################################################
    ##################################################
    ##################################################
    # computing summary metrics for segments
    if "segment_metrics" in outfiles:
        E.info( "computing summary metrics for segments" )
        outfile = outfiles["segment_metrics"]
        outfile.write( "track\tsection\tmetric\t%s\n" % "\t".join(Stats.Summary().getHeaders() ))
        for track in segments.tracks:
            IO.outputMetrics(  outfile,
                               segments[track], 
                               workspace,
                               track,
                               'segments',
                               )
        E.info( "wrote summary metrics for segments to %s" % str(outfile))


    ##################################################
    ##################################################
    ##################################################
    # collect observed counts from segments
    E.info( "collecting observed counts" )
    observed_counts = []
    for counter in counters:
        observed_counts.append( GatEngine.computeCounts( \
                counter = counter,
                aggregator = sum,
                segments = segments,
                annotations = annotations,
                workspace = workspace,
                workspace_generator = workspace_generator ) )
        
    ##################################################
    ##################################################
    ##################################################
    # sample and collect counts
    ##################################################
    E.info( "starting sampling" )

    if cache:
        E.info( "samples are cached in %s" % cache)
        samples = GatEngine.SamplesCached( filename = cache )
    elif sample_files:
        if not output_samples_pattern:
            raise ValueError( "require output_samples_pattern if loading samples from files" )
        # build regex
        regex = re.compile( re.sub("%s", "(\S+)", output_samples_pattern ) )
        E.info( "loading samples from %i files" % len(sample_files) )
        samples = SamplesFile( filenames = sample_files,
                               regex = regex ) 
    else:
        samples = GatEngine.Samples()

    sampled_counts = {}
    old_sampled_counts = {}
    
    counts = E.Counter()

    ntracks = len(segments.tracks)

    for ntrack, track in enumerate(segments.tracks):
        
        segs = segments[track]

        E.info( "sampling: %s: %i/%i" % (track, ntrack+1, ntracks))

        if output_samples_pattern and not sample_files:
            filename = re.sub("%s", track, output_samples_pattern )
            E.debug( "saving samples to %s" % filename)
            dirname = os.path.dirname( filename )
            if dirname and not os.path.exists( dirname ): os.makedirs( dirname )
            if filename.endswith(".gz"):
                samples_outfile = gzip.open( filename, "w" )                
            else:
                samples_outfile = open( filename, "w" )
        else:
            samples_outfile = None

        if workspace_generator.is_conditional:
            outer_sampler = ConditionalSampler( num_samples, 
                                                samples, 
                                                samples_outfile, 
                                                sampler,
                                                workspace_generator, 
                                                counters,
                                                outfiles,
                                                num_threads = num_threads )
        else:
            outer_sampler = UnconditionalSampler( num_samples, 
                                                  samples, 
                                                  samples_outfile, 
                                                  sampler,
                                                  workspace_generator, 
                                                  counters,
                                                  outfiles,
                                                  num_threads = num_threads )

        counts_per_track = outer_sampler.sample( track, counts, counters, segs, annotations, workspace, outfiles )

        if samples_outfile: samples_outfile.close()

        sampled_counts[track] = counts_per_track
        
        # old code, refactor into loop to save samples
        if 0:
            E.info( "sampling stats: %s" % str(counts))
            if track not in samples:
                E.warn( "no samples for track %s" % track )
                continue

            # clean up samples
            del samples[track]

    E.info( "sampling finished: %s" % str(counts) )

    ##################################################
    ##################################################
    ##################################################
    ## build annotator results
    ##################################################
    E.info( "computing PValue statistics" )

    annotator_results = list()

    counter_id = 0
    for counter, observed_count in zip( counters, observed_counts ):

        for track, r in observed_count.iteritems():
            for annotation, observed in r.iteritems():
                temp_segs, temp_annos, temp_workspace = workspace_generator( segments[track], 
                                                                             annotations[annotation], 
                                                                             workspace )

                # if reference is given, p-value will indicate difference
                # The test that track and annotation are present is done elsewhere
                if reference:
                    ref = reference[track][annotation]
                else:
                    ref = None
                
                annotator_results.append( GatEngine.AnnotatorResultExtended( \
                        track = track,
                        annotation = annotation,
                        counter = counter.name,
                        observed = observed,
                        samples = sampled_counts[track][counter_id][annotation],
                        track_segments = temp_segs,
                        annotation_segments = temp_annos,
                        workspace = temp_workspace,
                        reference = ref,
                        pseudo_count = pseudo_count ) )
        counter_id += 1

    ##################################################
    ##################################################
    ##################################################
    ## dump large table with counts
    ##################################################
    if output_counts_pattern:
        for counter in counters:
            name = counter.name
            filename = re.sub("%s", name, output_counts_pattern )
            
            E.info( "writing counts to %s" % filename )
            output = [ x for x in annotator_results if x.counter == name ]
            outfile = IOTools.openFile( filename, "w")
            outfile.write("track\tannotation\tobserved\tcounts\n" )
        
            for o in output:
                outfile.write( "%s\t%s\t%i\t%s\n" % \
                                   (o.track, o.annotation,
                                    o.observed,
                                    ",".join(["%i" % x for x in o.samples]) ) )
                
    return annotator_results

def fromCounts( filename ):
    '''build annotator results from a tab-separated table
    with counts.'''

    annotator_results = []

    with IOTools.openFile( filename, "r") as infile:

        E.info( "loading data")

        header = infile.readline()
        if not header== "track\tannotation\tobserved\tcounts\n":
            raise ValueError("%s not a counts file: got %s" % (infile, header) )
        
        for line in infile:
            track, annotation, observed, counts = line[:-1].split( "\t" )
            samples = numpy.array( map(float, counts.split(",")), dtype=numpy.float )
            observed = float(observed)
            annotator_results.append( GatEngine.AnnotatorResult( 
                track = track,
                annotation = annotation,
                counter = "na",
                observed = observed,
                samples = samples ) )

    return annotator_results



#!/usr/bin/env python
################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
gat-run - run the genomic annotation tool
=============================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script compares one or more genomic segments of interest 
against one more other genomic annotations.

Usage
-----

Example::

   python gat-run.py 
      --segment-file=segments.bed.gz 
      --workspace-file=workspace.bed.gz 
      --annotation-file=annotations_architecture.bed.gz  
 
Type::

   python gat-run.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os, sys, re, optparse, collections, types, glob, time
import numpy

import gat
import gat.Experiment as E
import gat.IOTools as IOTools
import gat.IO as IO
import gat.Stats as Stats
import GatSegmentList
import GatEngine

def fromSegments( options, args ):
    '''run analysis from segment files. 

    This is the most common use case.
    '''

    tstart = time.time()

    ##################################################
    ##################################################
    ##################################################
    ## build segments
    segments, annotations, workspaces, isochores = IO.buildSegments( options )

    E.info( "intervals loaded in %i seconds" % (time.time() - tstart) )

    ##################################################
    ##################################################
    ##################################################
    ## open various additional output files
    ##################################################
    outfiles = {}
    for section in ("sample", 
                    "segment_metrics",
                    "sample_metrics",
                    ):
        if section in options.output_stats or \
            "all" in options.output_stats or \
                len( [ x for x in options.output_stats if re.search( x, "section" ) ] ) > 0:
            outfiles[section] = E.openOutputFile(section)

    if 'sample_metrics' in outfiles:
        outfiles['sample_metrics'].write( "track\tsection\tmetric\t%s\n" % "\t".join(Stats.Summary().getHeaders() ))

    # filter segments by workspace
    workspace = IO.applyIsochores( segments, 
                                   annotations, 
                                   workspaces,
                                   options, 
                                   isochores,
                                   truncate_segments_to_workspace = options.truncate_segments_to_workspace,
                                   truncate_workspace_to_annotations = options.truncate_workspace_to_annotations,
                                   restrict_workspace = options.restrict_workspace )

    ##################################################
    ##################################################
    ##################################################
    ## check memory requirements
    counts = segments.countsPerTrack() 
    max_counts = max(counts.values())
    # previous algorithm: memory requirements if all samples are stored
    memory = 8 * 2 * options.num_samples * max_counts * len(workspace)

    ##################################################
    ##################################################
    ##################################################
    # initialize sampler
    if options.sampler == "annotator":
        sampler = GatEngine.SamplerAnnotator(
            bucket_size = options.bucket_size,
            nbuckets = options.nbuckets )
    elif options.sampler == "shift":
        sampler = GatEngine.SamplerShift( 
            radius = options.shift_expansion,
            extension = options.shift_extension )
    elif options.sampler == "segments":
        sampler = GatEngine.SamplerSegments()
    elif options.sampler == "local-permutation":
        sampler = GatEngine.SamplerLocalPermutation()
    elif options.sampler == "global-permutation":
        sampler = GatEngine.SamplerGlobalPermutation()
    elif options.sampler == "brute-force":
        sampler = GatEngine.SamplerBruteForce()
    elif options.sampler == "uniform":
        sampler = GatEngine.SamplerUniform()
        
    ##################################################
    ##################################################
    ##################################################
    # initialize counter
    counters = []
    for counter in options.counters:
        if counter == "nucleotide-overlap":
            counters.append( GatEngine.CounterNucleotideOverlap() )
        elif counter == "nucleotide-density":
            counters.append( GatEngine.CounterNucleotideDensity() )
        elif counter == "segment-overlap":
            counters.append( GatEngine.CounterSegmentOverlap() )
        elif counter == "annotations-overlap":
            counters.append( GatEngine.CounterAnnotationsOverlap() )
        elif counter == "segment-midoverlap":
            counters.append( GatEngine.CounterSegmentMidpointOverlap() )
        elif counter == "annotations-midoverlap":
            counters.append( GatEngine.CounterAnnotationsMidpointOverlap() )
        else:
            raise ValueError("unknown counter '%s'" % counter )

    ##################################################
    ##################################################
    ##################################################
    ## initialize workspace generator
    if options.conditional == "unconditional":
        workspace_generator = GatEngine.UnconditionalWorkspace()
    elif options.conditional == "cooccurance":
        workspace_generator = GatEngine.ConditionalWorkspaceCooccurance()
    elif options.conditional == "annotation-centered":
        if options.conditional_extension == options.conditional_expansion == None:
            raise ValueError( "please specify either --conditional-expansion or --conditional-extension" )
        workspace_generator = GatEngine.ConditionalWorkspaceAnnotationCentered( options.conditional_extension,
                                                                          options.conditional_expansion )
    elif options.conditional == "segment-centered":
        if options.conditional_extension == options.conditional_expansion == None:
            raise ValueError( "please specify either --conditional-expansion or --conditional-extension" )

        workspace_generator = GatEngine.ConditionalWorkspaceSegmentCentered( options.conditional_extension,
                                                                       options.conditional_expansion )
    else:
        raise ValueError("unknown conditional workspace '%s'" % options.conditional )


    ##################################################
    ##################################################
    ##################################################
    ## check if reference is compplete
    ##################################################
    if options.reference:
        reference = options.reference
        for track in segments.tracks:
            if track not in options.reference:
                raise ValueError("missing track '%s' in reference" % track )
            r = options.reference[track]
            for annotation in annotations.tracks:
                if annotation not in r:
                    raise ValueError("missing annotation '%s' in annotations for track='%s'" % (annotation, track ))


    ##################################################
    ##################################################
    ##################################################
    ## compute
    ##################################################
    annotator_results = gat.run( segments, 
                                 annotations, 
                                 workspace,
                                 sampler, 
                                 counters,
                                 workspace_generator = workspace_generator,
                                 num_samples = options.num_samples,
                                 cache = options.cache,
                                 outfiles = outfiles,
                                 output_counts_pattern = options.output_counts_pattern,
                                 output_samples_pattern = options.output_samples_pattern,
                                 sample_files = options.sample_files,
                                 conditional = options.conditional,
                                 conditional_extension = options.conditional_extension,
                                 reference = options.reference,
                                 pseudo_count = options.pseudo_count,
                                 num_threads = options.num_threads )

    return annotator_results


def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    parser = gat.buildParser( usage = globals()["__doc__"] )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    ##################################################
    description_header, descriptions, description_width = IO.readDescriptions( options )

    ##################################################
    size_pos, size_segment = GatSegmentList.getSegmentSize()
    E.debug( "sizes: pos=%i segment=%i, max_coord=%i" % (size_pos, size_segment, 2**(8 * size_pos )))

    ##################################################
    # set default counter
    if not options.counters:
        options.counters.append( "nucleotide-overlap" )

    ##################################################
    if options.output_tables_pattern != None: 
        if "%s" not in options.output_tables_pattern:
            raise ValueError( "output_tables_pattern should contain at least one '%s'")

    if options.output_samples_pattern != None: 
        if "%s" not in options.output_samples_pattern:
            raise ValueError( "output_samples_pattern should contain at least one '%s'")

    if options.output_counts_pattern != None: 
        if "%s" not in options.output_counts_pattern:
            raise ValueError( "output_counts_pattern should contain at least one '%s'")


    ##################################################
    # read fold changes that results should be compared with
    if options.null != "default":
        if not os.path.exists( options.null ):
            raise OSError( "file %s not found" % options.null )
        E.info( "reading reference results from %s" % options.null )
        options.reference = IO.readAnnotatorResults( options.null )
    else:
        options.reference = None

    if options.input_filename_counts:
        # use pre-computed counts
        annotator_results = GatEngine.fromCounts( options.input_filename_counts )

    elif options.input_filename_results:
        # use previous results (re-computes fdr)
        E.info( "reading gat results from %s" % options.input_filename_results )
        annotator_results = IO.readAnnotatorResults( options.input_filename_results )

    else:
        # do full gat analysis
        annotator_results = fromSegments( options, args )

    ##################################################
    if options.pvalue_method != "empirical":
        E.info("updating pvalues to %s" % options.pvalue_method )
        GatEngine.updatePValues( annotator_results, options.pvalue_method )

    ##################################################
    ## output
    IO.outputResults( annotator_results, 
                      options, 
                      GatEngine.AnnotatorResultExtended.headers,
                      description_header, 
                      description_width,
                      descriptions )

    IO.plotResults( annotator_results, options )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



#!/usr/bin/env python
################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
great - run great analysis
==========================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script compares one or more genomic segments of interest 
against one more other genomic annotations.

GREAT says that a segment overlaps an annotation if the midpoint
of the segment overlaps the annotation.

The GREAT analysis is done on a per-isochore basis and overall.
For the overall analysis, there is no isochore correction
as the union of all isochores is treated as a single workspace.

This script implies several counters (``--counter`` option):

binom
    binomial model (as implemented by GREAT). Returns the probability
    of a certain number of segments overlapping (mid-point overlap) 
    an annotation

hyperg
    hypergeometric model. Returns the probabibility of a certain
    number of nucleotides within segments overlapping an annotation.
    Conceptually, all genomic positions are treated independently and 
    segments are fragmented into a collection of single bases that are
    placed indepently. This is valid for point intervals such as SNPs,
    but of course ignores segment size, though it might be used as a
    guide towards the expected fold enrichment, that can be computed
    more accurately using :mod:`gat-run.py`

Usage
-----

Example::

   python gat-great.py 
      --segment-file=segments.bed.gz 
      --workspace-file=workspace.bed.gz 
      --annotation-file=annotations_architecture.bed.gz  
 
Type::

   python gat-run.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os, sys, re, optparse, collections, types, glob, time
import numpy
import numpy.random

import gat
import gat.Experiment as E
import gat.IOTools as IOTools
import gat.IO as IO
import GatSegmentList

import scipy.stats

GREAT_RESULT = collections.namedtuple( "GREAT",
                                       ("track",
                                        "annotation",
                                        "isochore",
                                        "counter",
                                        "observed", # nsegments_overlapping_annotations
                                        "expected",
                                        "nsegments_in_workspace",
                                        "nannotations_in_workspace",
                                        "nsegments_overlapping_annotation",
                                        "nannotations_overlapping_segments",
                                        "basecoverage_intersection",
                                        "basecoverage_segments",
                                        "basecoverage_annotation",
                                        "basecoverage_workspace",
                                        "fraction_coverage_annotation",
                                        "fold",
                                        "pvalue",
                                        "qvalue" ) )

def main( argv ):

    if not argv: argv = sys.argv

    # setup command line parser
    parser = optparse.OptionParser( version = "%prog version: $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-a", "--annotation-file", "--annotations", dest="annotation_files", type="string", action="append",
                      help="filename with annotations [default=%default]."  )

    parser.add_option("-s", "--segment-file", "--segments", dest="segment_files", type="string", action="append",
                      help="filename with segments. Also accepts a glob in parentheses [default=%default]."  )

    parser.add_option("-w", "--workspace-file", "--workspace", dest="workspace_files", type="string", action="append",
                      help="filename with workspace segments. Also accepts a glob in parentheses [default=%default]."  )

    parser.add_option("-i", "--isochore-file", "--isochores", dest="isochore_files", type="string", action="append",
                      help="filename with isochore segments. Also accepts a glob in parentheses [default=%default]."  )

    parser.add_option("-o", "--order", dest="output_order", type="choice",
                      choices = ( "track", "annotation", "fold", "pvalue", "qvalue" ),
                      help="order results in output by fold, track, etc. [default=%default]."  )

    parser.add_option("-q", "--qvalue-method", dest="qvalue_method", type="choice",
                      choices = ( "storey", "BH", "bonferroni", "holm", "hommel", "hochberg", "BY", "none" ),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default]."  )

    parser.add_option( "--qvalue-lambda", dest="qvalue_lambda", type="float",
                      help="fdr computation: lambda [default=%default]."  )

    parser.add_option( "--qvalue-pi0-method", dest="qvalue_pi0_method", type="choice",
                       choices = ("smoother", "bootstrap" ),
                       help="fdr computation: method for estimating pi0 [default=%default]."  )
    parser.add_option( "--descriptions", dest="input_filename_descriptions", type="string", 
                       help="filename mapping annotation terms to descriptions. "
                            " if given, the output table will contain additional columns "
                            " [default=%default]" )

    parser.add_option( "--ignore-segment-tracks", dest="ignore_segment_tracks", action="store_true", 
                       help="ignore segment tracks - all segments belong to one track [default=%default]" )

    parser.add_option( "--enable-split-tracks", dest="enable_split_tracks", action="store_true", 
                       help="permit the same track to be in multiple files [default=%default]" )

    parser.add_option( "--output-bed", dest="output_bed", type="choice", action="append",
                       choices = ( "all", 
                                   "annotations", "segments", 
                                   "workspaces", "isochores",
                                   "overlap" ),
                       help="output bed files [default=%default]."  )

    parser.add_option( "--output-stats", dest="output_stats", type="choice", action="append",
                       choices = ( "all", 
                                   "annotations", "segments", 
                                   "workspaces", "isochores",
                                   "overlap" ),
                       help="output overlap summary stats [default=%default]."  )

    parser.add_option( "--restrict-workspace", dest="restrict_workspace", action="store_true", 
                       help="restrict workspace to those segments that contain both track"
                       " and annotations [default=%default]" )

    parser.add_option( "--counter", dest="counters", type="choice", action="append",
                       choices = ( "binom", "hyperg" ),
                       help="counter to use [default=%default]."  )

    parser.add_option( "--output-tables-pattern", dest="output_tables_pattern", type="string", 
                      help="output pattern for result tables. Used if there are multiple counters used [default=%default]."  )

    parser.set_defaults(
        annotation_files = [],
        segment_files = [],
        workspace_files = [],  
        sample_files = [],
        counters = [],
        output_stats = [],
        output_bed = [],
        output_tables_pattern = "%s.tsv.gz",
        output_order = "fold",
        input_filename_counts = None,
        input_filename_results = None,
        pvalue_method = "empirical",
        output_plots_pattern = None,
        output_samples_pattern = None,
        qvalue_method = "storey",
        qvalue_lambda = None,
        qvalue_pi0_method = "smoother",
        ignore_segment_tracks = False,
        input_filename_descriptions = None,
        conditional = "unconditional",
        conditional_extension = None,
        conditional_expansion = None,
        restrict_workspace = False,
        enable_split_tracks = False,
        shift_expansion = 2.0,
        shift_extension = 0,
        overlap_mode = "midpoint",
        truncate_workspace_to_annotations = False,
        truncate_segments_to_workspace = False
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    tstart = time.time()

    if len(options.counters) == 0:
        options.counters.append("binom")

    ############################################
    segments, annotations, workspaces, isochores = IO.buildSegments( options )
    E.info( "intervals loaded in %i seconds" % (time.time() - tstart) )

    # filter segments by workspace
    workspace = IO.applyIsochores( segments, annotations, workspaces, options, isochores )

    ############################################
    description_header, descriptions, description_width = IO.readDescriptions( options )

    ############################################
    ############################################
    ## compute per contig

    # compute bases covered by workspace
    workspace2basecoverage, isochores = {}, []
    for contig, ww in workspace.iteritems():
        workspace2basecoverage[contig] = ww.sum()
        isochores.append( contig )
    
    # compute percentage of bases covered by annotations in workspace
    # per isochore
    annotation2basecoverage = collections.defaultdict( dict )
    for annotation, aa in annotations.iteritems():
        for isochore, a in aa.iteritems():
            # need to truncate to workspace?
            annotation2basecoverage[annotation][isochore] = a.sum()
            
    results_per_contig = collections.defaultdict( list )

    E.info( "computing counts per isochore" )
    # results per isochore 
    def emptyResult( segment, annotation, isochore, 
                     counter,
                     nsegments_in_workspace,
                     basecoverage_annotation,
                     basecoverage_workspace):
        return GREAT_RESULT._make( (
                segment, annotation, isochore, 
                counter,
                0, # observed
                0, # expected
                nsegments_in_workspace,
                0, # nannotations_in_workspace
                0, # nsegments_overlapping_annotation
                0, # nannotations_overlapping_segments
                0, # basecoverage_intersection
                0, # basecoverage_segments
                basecoverage_annotation,
                basecoverage_workspace,
                0.0,
                1.0,
                1.0,
                1.0 ))

    for isochore in isochores:    
        basecoverage_workspace = workspace2basecoverage[isochore]

        # iterate over all isochores        
        for segment, segmentdict in segments.iteritems():
            try:
                ss = segmentdict[isochore]
                # select segments overlapping workspace
                segments_in_workspace = GatSegmentList.SegmentList( clone = ss )
                segments_in_workspace.intersect( workspace[isochore] )
                # number of segments in workspace
                nsegments_in_workspace = len(segments_in_workspace)
            except KeyError:
                ss = None

            basecoverage_segments = segments_in_workspace.sum()

            for annotation, annotationdict in annotations.iteritems():
                
                # if annotation != "GO:0030957": continue

                try:
                    aa = annotationdict[isochore]
                except KeyError:
                    aa = None

                # p_A: proportion of bases covered by annotation
                try:
                    basecoverage_annotation = annotation2basecoverage[annotation][isochore]
                except KeyError:
                    basecoverage_annotation = 0

                if ss == None or aa == None:
                    for counter in options.counters:
                        results_per_contig[(counter,segment,annotation)].append( emptyResult(segment, annotation,
                                                                                             isochore, 
                                                                                             counter,
                                                                                             nsegments_in_workspace,
                                                                                             basecoverage_annotation,
                                                                                             basecoverage_workspace ) )
                    continue
                
                # select segments overlapping annotation
                segments_overlapping_annotation = GatSegmentList.SegmentList( clone = ss )
                segments_overlapping_annotation.intersect( annotations[annotation][isochore] )
                # number of segments in annotation
                nsegments_overlapping_annotation = ss.intersectionWithSegments( annotations[annotation][isochore],
                                                                                mode = options.overlap_mode )
                                         
                # number of nucleotides at the intersection of segments, annotation and workspace
                basecoverage_intersection = segments_overlapping_annotation.sum()

                annotations_overlapping_segments = GatSegmentList.SegmentList( clone = aa )
                annotations_overlapping_segments.intersect( ss )
                nannotations_overlapping_segments = len( annotations_overlapping_segments )

                nannotations_in_workspace = len( aa )
                if nannotations_in_workspace == 0: 
                    for counter in options.counters:
                        results_per_contig[(counter,segment,annotation)].append( emptyResult(segment, 
                                                                                             annotation, 
                                                                                             isochore, 
                                                                                             counter,
                                                                                             nsegments_in_workspace,
                                                                                             basecoverage_annotation,
                                                                                             basecoverage_workspace ) )
                    continue

                fraction_coverage_annotation = basecoverage_annotation / float( basecoverage_workspace )
                fraction_hit_annotation = float(nannotations_overlapping_segments) / nannotations_in_workspace
                
                for counter in options.counters:
                    if counter.startswith( "binom" ):
                        # GREAT binomial probability over "regions"
                        # n = number of genomic regions = nannotations_in_workspace
                        # ppi = fraction of genome annotated by annotation = fraction_coverage_annotation
                        # kpi = genomic regions with annotation hit by segments = nannotations_in_segments
                        # sf = survival functions = 1 -cdf
                        # probability of observing >kpi in a sample of n where the probabily of succes is
                        # ppi.
                        pvalue = scipy.stats.binom.sf( nsegments_overlapping_annotation - 1, 
                                                       nsegments_in_workspace, 
                                                       fraction_coverage_annotation )
                        
                        expected = fraction_coverage_annotation * nsegments_in_workspace
                        observed = nsegments_overlapping_annotation 

                    elif counter.startswith( "hyperg" ):
                        
                        # hypergeometric probability over nucleotides
                        # Sampling without replacement
                        # x,M,n,M
                        # x = observed number of nucleotides in overlap of segments,annotations and workspace
                        # M = number of nucleotides in workspace
                        # n = number of nucleotides in annotations (and workspace)
                        # N = number of nucleotides in segments (and workspace)
                        # P-value of obtaining >x number of nucleotides overlapping.
                        rv = scipy.stats.hypergeom( basecoverage_workspace,
                                                    basecoverage_annotation,
                                                basecoverage_segments )
                    
                        pvalue = rv.sf( basecoverage_intersection )
                        expected = rv.mean()
                        observed = basecoverage_intersection

                    if expected != 0:
                        fold = float(observed) / expected
                    else:
                        fold = 1.0

                    r = GREAT_RESULT._make( (
                            segment, annotation, isochore,
                            counter,
                            observed,
                            expected,
                            nsegments_in_workspace,
                            nannotations_in_workspace,
                            nsegments_overlapping_annotation,
                            nannotations_overlapping_segments,
                            basecoverage_intersection,
                            basecoverage_segments,
                            basecoverage_annotation,
                            basecoverage_workspace,
                            fraction_coverage_annotation,
                            fold,
                            pvalue,
                            1.0 ))
                    # print "\t".join( map(str, r))
                    results_per_contig[ (counter,segment,annotation) ].append( r )

    E.info( "merging counts per isochore" )

    # compute sums
    results = []
    
    for niteration, pair in enumerate(results_per_contig.iteritems()):

        counter, segment, annotation = pair[0]
        data = pair[1]
            
        nsegments_in_workspace = sum( [x.nsegments_in_workspace for x in data ] )
        nsegments_overlapping_annotation = sum( [x.observed for x in data ] )
        nannotations_in_workspace = sum( [x.nannotations_in_workspace for x in data ] )
        nannotations_overlapping_segments = sum( [x.nannotations_overlapping_segments for x in data ] )

        basecoverage_intersection = sum( [x.basecoverage_intersection for x in data ] )
        basecoverage_segments = sum( [x.basecoverage_segments for x in data ] )
        basecoverage_annotation = sum( [x.basecoverage_annotation for x in data ] )
        basecoverage_workspace = sum( [x.basecoverage_workspace for x in data ] )

        fraction_coverage_annotation = basecoverage_annotation / float( basecoverage_workspace )

        if counter.startswith( "binom" ):
            pvalue = scipy.stats.binom.sf( nsegments_overlapping_annotation-1, 
                                            nsegments_in_workspace, 
                                            fraction_coverage_annotation )
            expected = fraction_coverage_annotation * nsegments_in_workspace
            observed = nsegments_overlapping_annotation 
        elif counter.startswith( "hyperg" ):
            rv = scipy.stats.hypergeom( basecoverage_workspace,
                                        basecoverage_annotation,
                                        basecoverage_segments )
            
            pvalue = rv.sf( basecoverage_intersection )
            expected = rv.mean()
            observed = basecoverage_intersection


        if expected != 0:
            fold = float(observed) / expected
        else:
            fold = 1.0

        r = GREAT_RESULT._make( (
                segment, annotation, "all",
                counter,
                observed,
                expected,
                nsegments_in_workspace,
                nannotations_in_workspace,
                nsegments_overlapping_annotation,
                nannotations_overlapping_segments,
                basecoverage_intersection,
                basecoverage_segments,
                basecoverage_annotation,
                basecoverage_workspace,
                fraction_coverage_annotation,
                fold,
                pvalue,
                1.0 ))

        results.append( r )

    IO.outputResults( results, 
                      options, 
                      GREAT_RESULT._fields, 
                      description_header, 
                      description_width,
                      descriptions )

    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )

                            




"""test gat implementation."""

import unittest
import random, tempfile, shutil, os, re, gzip, sys
import numpy, math
import pickle

from GatEngine import AnnotatorResult, IntervalCollection, \
    Samples, SamplesCached, computeFDR, \
    SamplerAnnotator

from GatSegmentList import SegmentList

class GatTest( unittest.TestCase ):
    def shortDescription( self ): return None

class TestSegmentList( GatTest ):
    '''test segment list implementation.'''

    def testCreateAndClear( self ):
        s = SegmentList()
        self.assertEqual( 0, len(s) )
        s.add( 0, 100)
        self.assertEqual( 1, len(s) )
        s.clear()
        self.assertEqual( 0, len(s) )

    def testNormalize1( self ):
        '''non-overlapping segments.'''

        ss = [ (x, x + 10 ) for x in range( 0, 1000, 100) ]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss: s.add( start, end )
        s.normalize()

        self.assertEqual( len(s), 10 )
        self.assertEqual( s.sum(), 100 )
        s2 = SegmentList( iter = ss )
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testNormalize1b( self ):
        '''non-overlapping segments.'''

        ss = [ (x, x + 10 ) for x in range( 100, 1100, 100) ]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss: s.add( start, end )
        s.normalize()
        self.assertEqual( len(s), 10 )
        self.assertEqual( s.sum(), 100 )
        s2 = SegmentList( iter = ss )
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testNormalizeEmpty( self ):
        '''non-overlapping segments.'''

        s = SegmentList()
        self.assertEqual( len(s), 0)
        s.normalize()
        self.assertEqual( len(s), 0)
        self.assertEqual( s.isNormalized, 1)
        s2 = SegmentList()
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testNormalizeEmptySegment( self ):
        s = SegmentList( iter = [(0, 0),] )
        s.normalize()        
        self.assertEqual( s.isNormalized, 1)
        self.assertEqual( len(s), 0)

        s = SegmentList( iter = [(0, 0),(0,0)] )
        s.normalize()        
        self.assertEqual( s.isNormalized, 1)
        self.assertEqual( len(s), 0)

        ss = [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6), (0, 7), (0, 8), (0, 9)]
        s = SegmentList( iter = ss )
        s.normalize()        
        self.assertEqual( s.isNormalized, 1)
        self.assertEqual( len(s), 1)

        s2 = SegmentList( iter = ss )
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testNormalize2( self ):
        '''overlapping segments.'''

        ss = [ (x, x + 1000 ) for x in range( 0, 1000, 100) ]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss: s.add( start, end )
        s.normalize()
        self.assertEqual( len(s), 1 )
        self.assertEqual( s.sum(), 1900 )

    def testNormalize3( self ):
        '''non-overlapping but adjacent segments.'''

        ss = [ (x, x + 100 ) for x in range( 0, 1000, 100) ]
        random.shuffle(ss)
        s = SegmentList()
        for start, end in ss: s.add( start, end )
        s.normalize()

        self.assertEqual( len(s), 10 )
        self.assertEqual( s.sum(), 1000 )
        s2 = SegmentList( iter = ss )
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testNormalize4( self ):
        # test multiple interleaved segments
        ss = [ (x, x + 100 ) for x in range( 0, 1000, 10) ]
        s = SegmentList()
        for start, end in ss: s.add( start, end )
        s.normalize()
        self.assertEqual( len(s), 1 )
        self.assertEqual( s.sum(), 1090 )
        s2 = SegmentList( iter = ss )
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testNormalize5( self ):
        ss = [(489, 589), (1966, 2066), (2786, 2886), (0, 0), (3889, 3972), (3998, 4098), (6441, 6541), (6937, 7054), (7392, 7492), (8154, 8254), (9046, 9146)]
        s = SegmentList( iter = ss )
        s.normalize()
        self.assertEqual( len(s), len( [x for x in ss if x[1]-x[0] > 0] ) )
        self.assertEqual( s.sum(), 1000 )
        s2 = SegmentList( iter = ss )
        s2.merge( -1 )
        self.assertEqual( s, s2 )

    def testExtend( self ):
        
        s1 = SegmentList( iter =  [ (x, x + 100 ) for x in range( 0, 1000, 100) ] )
        s2 = SegmentList( iter =  [ (x, x + 100 ) for x in range( 2000, 3000, 100) ] )
        s1.extend(s2 )
        self.assertEqual( s1.sum(), s2.sum() * 2 )
        self.assertEqual( len(s1), len(s2) * 2 )

    def testTrim( self ):
        '''test trimming over full range of insertion points and deletions.'''

        for point in xrange( 0, 1000 ):
            for size in xrange( 0, 300 ):
                ss = [ (x, x + 100 ) for x in range( 0, 1000, 100) ]
                s = SegmentList( iter = ss,
                                     normalize = True )
                orig = s.sum()
                s.trim( point, size )
                self.assertEqual( orig - size, s.sum(),
                                  "trimming error at %i:%i: expected %i, got %i, %s" %\
                                      (point, size,
                                       orig-size, s.sum(), str(s) ) )


    def testInsertionPoint( self ):
        '''check insertion point for normalized segment lists'''
        ss = [ (x, x + 10 ) for x in range( 0, 100, 10) ]
        s = SegmentList( iter = ss,
                             normalize = True )
        
        for point in xrange( 0, 100 ):
            p = s.getInsertionPoint( point, point + 1 )
            self.assertEqual( p, point // 10 )

        ss = [ (x, x + 10 ) for x in range( 0, 100, 20) ]
        s = SegmentList( iter = ss,
                             normalize = True )
        
        for point in xrange( 0, 100 ):
            p = s.getInsertionPoint( point, point + 1 )
            if point >= 90: 
                self.assertEqual( p, len(s) )
            else:
                self.assertEqual( p, point // 20 ) 

        ss = [ (x, x + 10 ) for x in range( 10, 100, 20) ]
        s = SegmentList( iter = ss,
                             normalize = True )
        
        for point in xrange( 0, 100 ):
            p = s.getInsertionPoint( point, point + 1 )
            self.assertEqual( p, (point - 10) // 20 )

    def testInsertionPointNonNormalized( self ):
        '''check insertion point for unnormalized segment lists.'''
        ss = [ (x, x + 20 ) for x in range( 0, 100, 10) ]
        s = SegmentList( iter = ss,
                         normalize = False )
        
        for point in xrange( 0, 100 ):
            self.assertRaises( AssertionError, s.getInsertionPoint, point, point + 1 )

    def testMergeAdjacent( self ):
        ss = [ (x, x + 100  ) for x in range( 0, 1000, 100) ]
        random.shuffle(ss)
        s = SegmentList( iter = ss )
        s.merge( 0 )
        self.assertEqual( len(s), 1 )
        self.assertEqual( s.sum(), 1000 )

    def testMergeNeighbours( self ):
  
        for y in range( 0,5 ):
            ss = [ (x, x + 100 - y ) for x in range( 0, 1000, 100) ]
            random.shuffle(ss)
            for x in range(0,y+1):
                s = SegmentList( iter = ss )
                s.merge(  x )
                if x < y:
                    self.assertEqual( len(s), 10 )
                    self.assertEqual( s.sum(), 1000 - 10 * y)
                else:
                    self.assertEqual( len(s), 1 )
                    self.assertEqual( s.sum(), 1000 - y )

    def testExpand( self ):
        ss = [ (x, x + 10 ) for x in range( 0, 1000, 100) ]
        s = SegmentList( iter = ss, normalize = True )
        s.expand_segments( 2.0 )
        self.assertEqual( s.sum(), 195)
        self.assertEqual( len(s), 10 )

    def testLargest( self ):
        ss = [ (x, x + x / 10 ) for x in range( 0, 1000, 100) ]
        s = SegmentList( iter = ss, normalize = True )
        p = s.largest()
        self.assertEqual( p['end']-p['start'], 90)

    def testGetFilledSegmentsFromStart( self ):
        ss = [ (x, x + 10 ) for x in range( 0, 120, 20) ]
        s = SegmentList( iter = ss, normalize = True )
        for x in range( 0, 120, 5 ):
            f = s.getFilledSegmentsFromStart( x, 20 )
            self.assertEqual( f.sum(), 20 )
            if x >= 110:
                self.assertEqual( f.min(), s.min() )
                self.assertEqual( f.max(), 30 )
            elif x > 80: 
                self.assertEqual( f.min(), s.min() )
                self.assertEqual( f.max(), s.max() )
            else:
                if x in (0,20,40,60,80):
                    self.assertEqual( f.min(), x )
                    self.assertEqual( f.max(), x + 30 )
                elif x in (10,30,50,70):
                    self.assertEqual( f.min(), x + 10)
                    self.assertEqual( f.max(), x + 40 )
                elif x in (5,25,45,65):
                    self.assertEqual( f.min(), x )
                    self.assertEqual( f.max(), x + 40 )
                elif x in (15,35,55,75):
                    self.assertEqual( f.min(), x + 5 )
                    self.assertEqual( f.max(), x + 35 )
            self.assertEqual( s.sum(), s.getFilledSegmentsFromStart(x, 100 ).sum() )

    def testGetFilledSegmentsFromEnd( self ):
        ss = [ (x, x + 10 ) for x in range( 0, 120, 20) ]
        s = SegmentList( iter = ss, normalize = True )
        for x in range( 0, 120, 5 ):
            f = s.getFilledSegmentsFromEnd( x, 20 )
            self.assertEqual( f.sum(), 20 )
            if x == 0:
                self.assertEqual( f.max(), s.max() )
                self.assertEqual( f.min(), 80 )
            elif x < 30: 
                self.assertEqual( f.max(), s.max() )
                self.assertEqual( f.min(), s.min() )
            else:
                if x in (40,60,80,100):
                    self.assertEqual( f.min(), x - 40 )
                    self.assertEqual( f.max(), x - 10 )
                elif x in (30,50,70,90):
                    self.assertEqual( f.min(), x - 30 )
                    self.assertEqual( f.max(), x )
                elif x in (45,65,85):
                    self.assertEqual( f.min(), x - 40)
                    self.assertEqual( f.max(), x  )
                elif x in (35,55,75,95):
                    self.assertEqual( f.min(), x - 35 )
                    self.assertEqual( f.max(), x - 5 )
                    
            self.assertEqual( s.sum(), s.getFilledSegmentsFromEnd(x, 100 ).sum() )

    def testPickling( self ):
        ss = [ (x, x + 10 ) for x in range( 0, 120, 20) ]
        s = SegmentList( iter = ss, normalize = True )

        b = pickle.loads(pickle.dumps(s))
        
        self.assertEqual( s, b )

    def testSharing( self ):
        
        ss = [ (x, x + 10 ) for x in range( 0, 120, 20) ]
        s = SegmentList( iter = ss, normalize = True )
        s.share( "/testshare" )
        n = SegmentList( share = s )
        self.assertEqual( s, n )

    def testPickledSharing( self ):
        
        ss = [ (x, x + 10 ) for x in range( 0, 120, 20) ]
        s = SegmentList( iter = ss, normalize = True )
        s.share( "/testshare" )
        b = pickle.loads(pickle.dumps(s))
        self.assertEqual( s, b )

    def testUnsharing( self ):
        
        ss = [ (x, x + 10 ) for x in range( 0, 120, 20) ]
        s = SegmentList( iter = ss, normalize = True )
        s.share( "/testshare" )
        s.unshare()
        s.share( "/testshare" )
        s.unshare()

class TestSegmentListOverlap( GatTest ):
    
    def setUp( self ):
        self.a = SegmentList( iter = ( (x, x + 10 ) for x in range( 0, 1000, 100) ), normalize = True )

    def testOverlapFull( self ):
        self.assertEqual( self.a.overlapWithRange( 0, 1000), self.a.sum() )

    def testOverlapHalf( self ):
        self.assertEqual( self.a.overlapWithRange( 0, 500), self.a.sum() / 2)
        self.assertEqual( self.a.overlapWithRange( 500, 1000), self.a.sum() / 2)

    def testOverlapAfter( self ):
        self.assertEqual( self.a.overlapWithRange( 900, 910), 10 )
        self.assertEqual( self.a.overlapWithRange( 905, 915), 5 )

    def testOverlapNone( self ):
        self.assertEqual( self.a.overlapWithRange( 1000, 2000), 0 )
        self.assertEqual( self.a.overlapWithRange( 2000, 3000), 0 )

    def testOverlapOutOfRange( self ):
        self.assertRaises( OverflowError, self.a.overlapWithRange, -100, 5 )
        self.assertRaises( OverflowError, self.a.overlapWithRange, -100, -50 )
        self.assertEqual( self.a.overlapWithRange( 905,1100 ), 5)
        self.assertEqual( self.a.overlapWithRange( 1000,1100 ), 0)

    def testOverlapAll( self ):
        for x in range( 0, 1000, 100):
            for y in range( 0, 10 ):
                self.assertEqual( self.a.overlapWithRange( x+y,x+y+1), 1, \
                                      "no overlap failure at %i: %i" % (x+y, self.a.overlapWithRange( x+y,x+y+1)))
            for y in range( 10, 100 ):
                self.assertEqual( self.a.overlapWithRange( x+y,x+y+1), 0, "overlap failure at %i" % (x+y) )


class TestSegmentListIntersection( GatTest):

    def setUp( self ):
        #[(0, 10), (100, 110), (200, 210), (300, 310), (400, 410), (500, 510), (600, 610), (700, 710), (800, 810), (900, 910)]
        self.a = SegmentList( iter = ( (x, x + 10 ) for x in range( 0, 1000, 100) ), normalize = True )

    def testIntersectionFull( self ):
        b = SegmentList( iter = [ (0, 1000) ], normalize = True  ) 
        r = b.intersect( self.a )
        self.assertEqual( r.asList(), self.a.asList() )

    def testIntersectionSelf( self ):
        r = self.a.intersect( self.a )
        self.assertEqual( r.asList(), self.a.asList() )

    def testIntersectionCopy( self ):
        b = SegmentList( clone = self.a )
        r = b.intersect( self.a )
        self.assertEqual( r.asList(), self.a.asList() )
        
    def testNoIntersection( self ):
        b = SegmentList( iter = ( (x, x + 10 ) for x in range( 10, 1000, 100) ), normalize = True )
        r = b.intersect( self.a )
        self.assertEqual( r.asList(), [] )
        self.assertEqual( r.isEmpty, True )

    def testPartialIntersection( self ):
        b = SegmentList( iter = ( (x, x + 10 ) for x in range( 5, 1000, 100) ), normalize = True )
        r = b.intersect( self.a )
        self.assertEqual( len(r), len(self.a) )
        self.assertEqual( r.sum(), self.a.sum() / 2 )

    def testOverlap( self ):
        '''test if number of segments intersection is correct.'''

        b = SegmentList( iter = ( (x, x + 10 ) for x in range( 5, 1000, 100) ), normalize = True )
        self.assertEqual( self.a.intersectionWithSegments( b ), len(b) )
        self.assertEqual( b.intersectionWithSegments( self.a ), len(b) )

        # no intersection
        b = SegmentList( iter = ( (x, x + 10 ) for x in range( 10, 1000, 100) ), normalize = True )
        self.assertEqual( self.a.intersectionWithSegments( b ), 0 )
        self.assertEqual( b.intersectionWithSegments( self.a ), 0 )
        
        # double the number of segments in b
        b = SegmentList( iter = [(x, x + 5 ) for x in range( 0, 1000, 100) ] +\
                                 [(x+5, x + 10 ) for x in range( 0, 1000, 100) ], \
                                 normalize = True )
        self.assertEqual( self.a.intersectionWithSegments( b ), 10 )
        self.assertEqual( b.intersectionWithSegments( self.a ), 20 )

    def testFilter( self ):
        
        b = SegmentList( iter = ( (x, x + 5 ) for x in range( 500, 2000, 100) ), normalize = True )
        self.assertEqual( b.filter(self.a).asList(), [(500, 505), (600, 605), (700, 705), (800, 805), (900, 905)] )

        b = SegmentList( iter = ( (0,56), ) )
        c = SegmentList( iter = [(0, 50), (75, 125)] )
        self.assertEqual( b.filter(c).asList(), [(0,56)] )

        b = SegmentList( iter = ( (0,56), ) )
        c = SegmentList( iter = [(0, 10)] )
        self.assertEqual( b.filter(c).asList(), [(0,56)] )

class TestSegmentListSubtract( GatTest):

    def setUp( self ):
        #[(0, 10), (100, 110), (200, 210), (300, 310), (400, 410), (500, 510), (600, 610), (700, 710), (800, 810), (900, 910)]
        self.a = SegmentList( iter = ( (x, x + 10 ) for x in range( 0, 1000, 100) ), normalize = True )

    def testCompleteOverlap( self ):
        b = SegmentList( iter = [ (0, 1000) ], normalize = True  ) 
        r = b.subtract( self.a )
        c = [(10L, 100L), (110L, 200L), (210L, 300L), (310L, 400L), (410L, 500L), (510L, 600L), (610L, 700L), (710L, 800L), (810L, 900L), (910L, 1000L)]
        self.assertEqual( r.asList(), c )

    def testFullSubtraction( self ):
        b = SegmentList( iter = [ (0, 1000) ], normalize = True  ) 
        r = self.a.subtract( b )
        self.assertEqual( len(r), 0 )

    def testSelfSubtraction( self ):
        r = self.a.subtract( self.a )
        self.assertEqual( len(r), 0)

    def testSameSubtraction( self ):
        b = SegmentList( clone = self.a )
        r = b.subtract( self.a )
        self.assertEqual( len(r), 0)

    def testOverlap( self ):

        b = SegmentList( iter = ( (x, x + 10 ) for x in range( 5, 1000, 100) ), normalize = True )
        r = b.subtract( self.a )
        c = [(10L, 15L), (110L, 115L), (210L, 215L), (310L, 315L), (410L, 415L), (510L, 515L), (610L, 615L), (710L, 715L), (810L, 815L), (910L, 915L)]
        self.assertEqual( r.asList(), c )


    def testSingleSegmentSubtraction( self ):
        a = SegmentList( iter = [ (0,12000)], normalize = True )
        b = SegmentList( iter = [ (0,10000)], normalize = True )
        r = a.subtract( b )
        self.assertEqual( r.asList(), [(10000,12000)] )

class TestIntervalCollection( GatTest):

    def setUp( self ):
        self.a = IntervalCollection( "a" )
        self.a.add( "track1", "contig1",
                    SegmentList( iter = ( (x, x + 10 ) for x in range( 0, 1000, 100) ), normalize = True ) )
        self.a.add( "track1", "contig2",
                    SegmentList( iter = ( (x, x + 10 ) for x in range( 0, 1000000, 100) ), normalize = True ) )
        self.a.add( "track2", "contig1",
                    SegmentList( iter = ( (x, x + 10 ) for x in range( 1000, 2000, 100) ), normalize = True ) )
        
    def testSaveLoad( self ):
        
        fn = "tmp_testSaveLoad.bed"
        outfile = open(fn, "w")
        self.a.save( outfile )
        outfile.close()

        b = IntervalCollection("b")
        b.load( fn )

        self.assertEqual( len(self.a), len(b) )
        self.assertEqual( self.a.tracks, b.tracks )
        self.assertEqual( self.a.sum(), b.sum() )
        for t in b.tracks:
            self.assertEqual( sorted(self.a[t].keys()), sorted(b[t].keys()) )
            for x in self.a[t].keys():
                self.assertEqual( self.a[t][x], b[t][x] )

    def testSharing( self ):

        aa = self.a.clone()
        aa.share()

class TestToFromIsochores( GatTest ):

    def setUp( self ):
        
        a = IntervalCollection( "a" )

        # every ten at 
        a.add( "track1", "contig1",
               SegmentList( iter = ( (x, x + 250 ) for x in range( 0, 2000, 500) ), normalize = True ) )

        a.add( "track1", "contig2",
               SegmentList( iter = ( (x, x + 250 ) for x in range( 0, 2000, 500) ), normalize = True ) )

        self.a = a

    def check( self, a, orig ):
        
        self.assertEqual( a.tracks, orig.tracks )
        self.assertEqual( sorted(a["track1"].keys()), 
                          sorted( ['contig2', 'contig1']))

        for contig in ("contig1", "contig2"):
            self.assertEqual( list(a["track1"][contig]), list(orig["track1"][contig]) )
    
    def testNoIsochores( self ):
        
        a = self.a.clone()
        orig = a.clone()
        a.fromIsochores()
        self.check( a, orig )

    def testToFromIsochores( self ):

        a = self.a.clone()
        orig = a.clone()

        isochores = IntervalCollection( "isochores" )

        # covering isochores of size 100
        isochores.add( "highGC", "contig1",
                       SegmentList( iter = ( (x, x + 100 ) for x in range( 0, 10000, 200) ), normalize = True ) )

        isochores.add( "lowGC", "contig1",
                       SegmentList( iter = ( (x, x + 100 ) for x in range( 100, 10000, 200) ), normalize = True ) )

        isochores.add( "highGC", "contig2",
                       SegmentList( iter = ( (x, x + 100 ) for x in range( 0, 10000, 200) ), normalize = True ) )

        isochores.add( "lowGC", "contig2",
                       SegmentList( iter = ( (x, x + 100 ) for x in range( 100, 10000, 200) ), normalize = True ) )

        a.toIsochores( isochores )
        self.assertEqual( a.tracks, orig.tracks )
        self.assertEqual( sorted(a["track1"].keys()), 
                          sorted( ['contig2.highGC', 'contig1.highGC', 'contig2.lowGC', 'contig1.lowGC']))
        
        a.fromIsochores()

        self.check( a, orig )

class TestPValue( GatTest ):
    '''test if pvalue computation is correct.'''
    
    def testPValue1( self ):
        
        observed = 0.332640195285
        values = [0.3593727449353678, 0.24446041723385858, 0.11321078358680142, 0.28500665546177717, 0.017634423032620888, 0.47144573882791929, 0.20295266762886535, 0.24374906675401431, 0.12536987767373536, 0.36647407597049514, 0.1317950839045143, 0.32036858313479905, 0.2131875486832529, 0.18211958887292382, 0.4382662865088186, 0.12487068923091568, 0.38895983423268921, 0.43156050120631062, 0.18784825518278428, 0.23958644581530344, 0.16386055449534453, 0.42697777787951602, 0.07748674945294963, 0.47881248869131277, 0.37267534771232319, 0.8083924735050152, 0.29179189019925428, 0.29802029242077777, 0.2054027587360118, 0.10766996738143179, 0.39134998593956405, 0.36412616130029274, 0.37015995608450686, 0.61246049427537563, 0.59897086243095388, 0.20718454055122912, 0.14334918487088333, 0.42189815231899974, 0.21738749430714899, 0.39304902005163428, 0.50261637732761, 0.20759334134444557, 0.21005124432686503, 0.31027042275886835, 0.71335371670327341, 1.4192781030245714, 0.50672517580861098, 0.18067653694488042, 0.85952730574991043, 0.19249388587333111, 0.18826477050167958, 0.22742885130411533, 0.24125995809534906, 0.045750800392306591, 0.78242626285998884, 0.20614461737324383, 0.56783904985512668, 0.33500622312674566, 0.043533317315170454, 0.27874382197104552, 0.3685525858770754, 0.1751812314517863, 0.2532293526642409, 0.15785104775566922, 0.2390711833181299, 0.42911409505776471, 0.16819203200742916, 0.40372196988518594, 0.43241512178368696, 0.30424021778439686, 0.19085162018033855, 0.58462246847853661, 0.631050399423982, 0.30137309454374051, 0.27565096287611918, 0.33033553618821287, 0.47665164689105288, 0.34084703029218633, 0.27978844627773986, 0.010536582324145049, 0.050935298127348511, 0.23536721808983668, 0.22364077067346355, 0.31704429093519465, 1.0296141286403104, 0.38123158028252929, 0.27538594123938104, 0.81446088474774558, 0.2660327021825486, 0.195234318277725, 0.462999083371401, 1.0587870384537303, 0.40260375543813692, 0.39471961997665139, 0.29845505700189406, 1.0259474557694457, 0.52111381852233729, 0.29182304834212835, 0.34045457181768657, 0.20417518807825608]
        
        result = AnnotatorResult("test", "test", "na", observed, values, reference = None, pseudo_count = 0 )

        self.assertEqual( result.pvalue, 0.57 )
        
class TestSamples( GatTest ):
    
    nsamples = 1000
    ntracks = 50
    nsegments = 10000
    nisochores = 20

    def testDelete( self ):
        '''test to track down a memory leak.'''
        # from guppy import hpy
        # hp = hpy()
        # hp.setrelheap()
        return

        samples = Samples()

        for track in xrange(self.ntracks):
            track_id = str(track)
            # print track_id
            for sample in xrange(self.nsamples):
                sample_id = str(sample)
                for isochore in xrange( self.nisochores):
                    isochore_id = str(isochore)
                    r = SegmentList( allocate = self.nsegments )
                    samples.add( track_id, sample_id, isochore_id, r )

            del samples[track_id]
            # print len(samples)
            # h = hp.heap()

            # print h

class TestCaching( GatTest ):

    sample_size = 10
    workspace_size = 1000

    def testCaching( self ):

        workspaces, segments, annotations = \
            IntervalCollection( "workspace" ), \
            IntervalCollection( "segment" ), \
            IntervalCollection( "annotation" )

        workspaces.add( "default", "chr1", SegmentList( iter = [(0,self.workspace_size),],
                                                            normalize = True ) )
        workspace = workspaces["default"]

        segments.add( "default", "chr1", SegmentList( iter = [(0,1),],
                                                          normalize = True ) )

        # annotations: a collection of segments with increasing density
        # all are overlapping the segments
        for y in range(1, 100, 2 ):
            annotations.add( "%03i" % y, "chr1",
                             SegmentList( iter = [(0,y),],
                                          normalize = True ) ) 
            
        workspace_size = workspace["chr1"].sum()

        sampler = SamplerAnnotator( bucket_size = 1, nbuckets = self.workspace_size )
        
        if os.path.exists( "test.cache"): os.remove( "test.cache" )
        
        outsamples = SamplesCached( "test.cache" )
        saved_samples = {}

        for track in segments.tracks:
            segs = segments[track]
            for x in xrange( self.sample_size ):
                for isochore in segs.keys():
                    r = sampler.sample( segs[isochore], workspace[isochore] )
                    saved_samples[(track,x,isochore)] = r
                    outsamples.add( track, x, isochore, r )
        
        del outsamples
        
        insamples = SamplesCached( "test.cache" )
        
        for track in segments.tracks:
            segs = segments[track]
            for x in xrange( self.sample_size ):
                for isochore in segs.keys():
                    insamples.load( track, x, isochore )
                    
        # compare
        for track in segments.tracks:
            segs = segments[track]
            for x in xrange( self.sample_size ):
                for isochore in segs.keys():
                    self.assertEqual( saved_samples[(track,x,isochore)].asList(),
                                      insamples[track][x][isochore].asList() )

        if os.path.exists( "test.cache"): os.remove( "test.cache" )
        if os.path.exists( "test.cache.idx"): os.remove( "test.cache.idx" )

class TestStats( GatTest ):

    ntracks = 10 # 17
    nannotations = 10 # 90
    nsamples = 1000

    def testPValueComputation( self ):
        
        annotation_size = 100
        workspace_size = 1000
        segment_size = 10

        l = 10

        for y in xrange(1, l):

            samples = [1] * y + [0] * (l - y)

            for x, s in enumerate(samples):

                g = AnnotatorResult( "track", 
                                         "samples",
                                         'counter',
                                         s,
                                         samples )
                self.assertEqual( g.isSampleSignificantAtPvalue( x, g.pvalue ), True )
                                  

                t = 0
                for z, s2 in enumerate(samples):
                    t += g.isSampleSignificantAtPvalue( z, g.pvalue )
                fpr = float(t) / l

                # == should work, but there is a problem 
                # for pvalue = 0.5
                self.assert_( fpr >= g.pvalue - 0.0001, 
                              "fpr (%f) != pvalue (%f): y=%i, x=%i, s=%i, t=%i" % \
                                  ( fpr, g.pvalue, y, x, s, t) )


    def testPValueComputation2( self ):
        
        samples = [0] * 66 + [1] * 2 + [2] * 20 + [3] * 1 + [4] * 6 + [6] * 2 + [8] * 2 + [16] * 1
        
        obs = 16

        g = AnnotatorResult( "track", "samples",
                                 'counter',
                                 obs,
                                 samples )

        self.assertEqual( g.pvalue, 0.01 )


    def testStats( self ):
        
        annotation_size = 100
        workspace_size = 1000
        segment_size = 10

        observed = numpy.random.hypergeometric(annotation_size, 
                                               workspace_size - annotation_size,
                                               segment_size, 
                                               self.ntracks * self.nannotations )

        results = []
        x = 0
        for track in range(self.ntracks ):
            for annotation in range(self.nannotations ):
                samples = numpy.random.hypergeometric(annotation_size, 
                                                      workspace_size - annotation_size,
                                                      segment_size, 
                                                      self.nsamples )

                samples.sort()
                # print "obs", observed[x], "sample=", samples

                results.append( AnnotatorResult( str(track),     
                                                     str(annotation),
                                                     'counter',
                                                     observed[x],
                                                     samples ) )

                x += 1

        computeFDR( results )
        for r in results: 
            self.assert_( r.qvalue > 0.5, "%f" % r.qvalue  )

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
################################################################################
#
#   MRC FGU Computational Genomics Group
#
#   $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $
#
#   Copyright (C) 2009 Andreas Heger
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#################################################################################
'''
gat-compare - compare two gat runs
==================================

:Author: Andreas Heger
:Release: $Id$
:Date: |today|
:Tags: Python

Purpose
-------

This script compares gat runs and compares statistical significance of fold change differences between two or more
regions of interest.

This script requires counts files created by gat (option ``--output-counts-file``).

The reported observed value between sets 1 and 2 is the difference of fc2 and fc1:

   * observed = fold2 - fold1

The script accepts one or more output files from a gat run. 

If only a single file is given, the significance of fold change differences are computed 
for each pairwise comparison of annotations. Thus, the script might answer the question 
if a transcription factor is more enriched in promotors than in enhancers.

If multiple files are given, the script computes the fold change differences for the same 
annotation for all pairwise combinations of :term:`segments of interest` across the different
files. Thus, the script might answer the question if transcription factor A is more enriched
in promotors than transcription factor B is enriched in promotors.

Usage
-----

Example::

   python gat-compare.py tfA.counts.tsv.gz tfB.counts.tsv.gz
      
Type::

   python gat-compare.py --help

for command line help.

Documentation
-------------

Code
----

'''

import os, sys, re, optparse, collections, types, glob, time, math
import itertools
import numpy

import gat
import gat.Experiment as E
import gat.IOTools as IOTools
import gat.IO as IO
import GatEngine

def main( argv = None ):
    """script main.

    parses command line options in sys.argv, unless *argv* is given.
    """

    if not argv: argv = sys.argv

    # setup command line parser
    parser = optparse.OptionParser( version = "%prog version: $Id: script_template.py 2871 2010-03-03 10:20:44Z andreas $", 
                                    usage = globals()["__doc__"] )

    parser.add_option("-o", "--order", dest="output_order", type="choice",
                      choices = ( "track", "annotation", "fold", "pvalue", "qvalue", "observed" ),
                      help="order results in output by fold, track, etc. [default=%default]."  )

    parser.add_option("-p", "--pvalue-method", dest="pvalue_method", type="choice",
                      choices = ( "empirical", "norm", ),
                      help="type of pvalue reported [default=%default]."  )

    parser.add_option("-q", "--qvalue-method", dest="qvalue_method", type="choice",
                      choices = ( "storey", "BH", "bonferroni", "holm", "hommel", "hochberg", "BY", "none" ),
                      help="method to perform multiple testing correction by controlling the fdr [default=%default]."  )

    parser.add_option( "--qvalue-lambda", dest="qvalue_lambda", type="float",
                      help="fdr computation: lambda [default=%default]."  )

    parser.add_option( "--qvalue-pi0-method", dest="qvalue_pi0_method", type="choice",
                       choices = ("smoother", "bootstrap" ),
                       help="fdr computation: method for estimating pi0 [default=%default]."  )

    parser.add_option( "--descriptions", dest="input_filename_descriptions", type="string", 
                       help="filename mapping annotation terms to descriptions. "
                            " if given, the output table will contain additional columns "
                            " [default=%default]" )

    parser.add_option( "--pseudo-count", dest="pseudo_count", type="float",
                      help="pseudo count. The pseudo count is added to both the observed and expected overlap. "
                       " Using a pseudo-count avoids gat reporting fold changes of 0 [default=%default]."  )

    parser.add_option( "--output-plots-pattern", dest="output_plots_pattern", type="string", 
                       help="output pattern for plots [default=%default]" )

    parser.set_defaults(
        pvalue_method = "empirical",
        qvalue_method = "BH",
        qvalue_lambda = None,
        qvalue_pi0_method = "smoother",
        # pseudo count for fold change computation to avoid 0 fc
        pseudo_count = 1.0,
        output_order = "observed",
        )

    ## add common options (-h/--help, ...) and parse command line 
    (options, args) = E.Start( parser, argv = argv, add_output_options = True )

    input_filenames_counts = args

    ##################################################
    E.info( "received %i filenames with counts" % len(input_filenames_counts))

    ##################################################
    description_header, descriptions, description_width = IO.readDescriptions( options )

    all_annotator_results = []

    for input_filename_counts in input_filenames_counts:
 
        E.info("processing %s" % input_filename_counts )
        
        annotator_results = gat.fromCounts( input_filename_counts )

        ##################################################
        if options.pvalue_method != "empirical":
            E.info("updating pvalues to %s" % options.pvalue_method )
            GatEngine.updatePValues( annotator_results, options.pvalue_method )

        ##################################################
        ##################################################
        ##################################################
        ## compute global fdr
        ##################################################
        E.info( "computing FDR statistics" )
        GatEngine.updateQValues( annotator_results, 
                           method = options.qvalue_method,
                           vlambda = options.qvalue_lambda,
                           pi0_method = options.qvalue_pi0_method )

        all_annotator_results.append( annotator_results )
        
    pseudo_count = options.pseudo_count
    results = []

    if len(all_annotator_results) == 1:
        E.info("performing pairwise comparison within a file")

        # collect all annotations
        annotations, segments = list(), set()
        for x in all_annotator_results[0]:
            segments.add( x.track)
            annotations.append( x )

        if len(segments) != 1:
            raise NotImplementedError( "multiple segments of interest")

        for data1, data2 in itertools.combinations(annotations, 2):
            
            # note that fold changes can be very large if there are 0 samples
            # this is fine for getting the distributional params (mean, stddev)
            fold_changes1 = data1.observed / (data1.samples + pseudo_count ) 
            fold_changes2 = data2.observed / (data2.samples + pseudo_count ) 

            # add a separate fc pseudo-count to avoid 0 values
            fold_changes1 += 0.0001
            fold_changes2 += 0.0001

            # Test is if relative fold change rfc is different from 1
            # note: rfc = fc1 / fc2 = obs1 / exp1 * obs2 / exp2
            #                       = obs1 / obs2 * exp2 / exp1
            # Thus, it is equivalent to test rfc = obs1/obs2 versus exp2 / exp1
            #
            # Convert to log space for easier plotting
            # Move the observed fold ratio in order to get an idea of the magnitude
            # of the underlying fold change
            delta_fold = data2.fold - data1.fold
            sampled_delta_fold = numpy.log( fold_changes1 / fold_changes2) + delta_fold
            observed_delta_fold = 0.0 + delta_fold

            result = GatEngine.AnnotatorResult( data1.annotation, data2.annotation, 
                                                "na",
                                                observed_delta_fold,
                                                sampled_delta_fold, 
                                                reference = None,
                                                pseudo_count = 0 )


            results.append( result )


    else:
        E.info("performing pairwise comparison between files")

        ##################################################
        # perform pairwise comparison
        for index1, index2 in itertools.combinations( range(len(input_filenames_counts)), 2):
            E.info( "comparing %i and %i" % (index1,index2))
            a,b = all_annotator_results[index1], all_annotator_results[index2]

            # index results in a and b
            aa = collections.defaultdict( dict )
            for x in a: aa[x.track][x.annotation] = x

            bb = collections.defaultdict( dict )
            for x in b: bb[x.track][x.annotation] = x

            if len(aa.keys()) != 1 or len(bb.keys()) != 1:
                raise NotImplementedError( "multiple segments of interest")

            track = "merged"
            # get shared annotations
            annotations1 = aa[track].keys()
            annotations2 = bb[track].keys()
            shared_annotations = list( set(annotations1).intersection( set(annotations2) ) )
            E.info( "%i shared annotations" % len(shared_annotations))

            for annotation in shared_annotations:

                # if not annotation.startswith("Ram:"): continue

                data1 = aa[track][annotation]
                data2 = bb[track][annotation]

                # note that fold changes can be very large if there are 0 samples
                # this is fine for getting the distributional params (mean, stddev)
                fold_changes1 = data1.observed / (data1.samples + pseudo_count ) 
                fold_changes2 = data2.observed / (data2.samples + pseudo_count ) 

                # add a separate fc pseudo-count to avoid 0 values
                fold_changes1 += 0.0001
                fold_changes2 += 0.0001

                # Test is if relative fold change rfc is different from 1
                # note: rfc = fc1 / fc2 = obs1 / exp1 * obs2 / exp2
                #                       = obs1 / obs2 * exp2 / exp1
                # Thus, it is equivalent to test rfc = obs1/obs2 versus exp2 / exp1
                #
                # Convert to log space for easier plotting
                # Move the observed fold ratio in order to get an idea of the magnitude
                # of the underlying fold change
                delta_fold = data2.fold - data1.fold
                sampled_delta_fold = numpy.log( fold_changes1 / fold_changes2) + delta_fold
                observed_delta_fold = 0.0 + delta_fold

                result = GatEngine.AnnotatorResult( track, annotation, 
                                                    "na",
                                                    observed_delta_fold,
                                                    sampled_delta_fold, 
                                                    reference = None,
                                                    pseudo_count = 0 )
                

                results.append( result )


    if len(results) == 0:
        E.critical( "no results found" )
        E.Stop()
        return
            
    IO.outputResults( results, 
                      options, 
                      GatEngine.AnnotatorResult.headers,
                      description_header, 
                      description_width,
                      descriptions,
                      format_observed = "%6.4f" )

    IO.plotResults( results, options )

    ## write footer and output benchmark information.
    E.Stop()

if __name__ == "__main__":
    sys.exit( main( sys.argv) )



'''test script to check memory usage in multiprocessing.'''


import multiprocessing
import GatSegmentList
import time

nsegments=100000000
nsegments=10000000
ncpu = 2
nwork = 100

s = GatSegmentList.SegmentList( iter = [(x,x+1) for x in xrange( 0,nsegments, 2)], normalize = True )

print "built list"
r = raw_input("press return")

s.share("/test")

print "shared data"
r = raw_input("press return")

def dowork( segs ):
    while 1: pass
    return segs.sum()

p = multiprocessing.Pool( ncpu )


print "starting mp"
r = p.map( dowork, [ s for i in range( nwork ) ] )

print r
